import pytest

from synrbl.mcs_search import MCSSearch
from synrbl.SynProcessor import CheckCarbonBalance

ID_COL = "id"
SOLVED_COL = "solved"
MCS_DATA_COL = "mcs"
REACTION_COL = "reaction"
REACTANTS_COL = "reactants"
PRODUCTS_COL = "products"
C_BALANCE_CHECK_COL = "carbon_balance_check"


def _init_mcs():
    return MCSSearch(ID_COL, solved_col=SOLVED_COL, mcs_data_col=MCS_DATA_COL)


def _add(data, reaction, solved=False):
    r_token = reaction.split(">>")
    entry = {
        ID_COL: len(data),
        REACTION_COL: reaction,
        SOLVED_COL: solved,
        REACTANTS_COL: r_token[0],
        PRODUCTS_COL: r_token[1],
    }
    check = CheckCarbonBalance(
        [entry], rsmi_col=REACTION_COL, symbol=">>", atom_type="C"
    )
    entry[C_BALANCE_CHECK_COL] = check.check_carbon_balance()[0]["carbon_balance_check"]
    data.append(entry)


def test_simple_mcs():
    mcs = _init_mcs()
    data = []
    _add(data, "COC(C)=O>>OC(C)=O")

    results = mcs.find(data)

    assert 1 == len(results)
    result = results[0]
    assert MCS_DATA_COL in result.keys()
    result = result[MCS_DATA_COL]

    assert 1 == len(result["smiles"])
    assert "C" == result["smiles"][0]

    assert 1 == len(result["boundary_atoms_products"])
    assert 1 == len(result["boundary_atoms_products"][0])
    assert {"C": 0} == result["boundary_atoms_products"][0][0]

    assert 1 == len(result["nearest_neighbor_products"])
    assert 1 == len(result["nearest_neighbor_products"][0])
    assert {"O": 1} == result["nearest_neighbor_products"][0][0]

    # assert 0 == len(result["issue"])
    # assert result["Certainty"] is True

    assert 1 == len(result["sorted_reactants"])
    assert "COC(C)=O" == result["sorted_reactants"][0]

    assert 1 == len(result["mcs_results"])
    assert "[#8]-&!@[#6](-&!@[#6])=&!@[#8]" == result["mcs_results"][0]


@pytest.mark.parametrize(
    "rxn, exp_smiles, exp_ba, exp_nn, exp_issue, exp_sr, exp_mcs",
    [
        (
            (
                "Nc1nc2c(ncn2C2OC(COP(=O)([O-])OP(=O)([O-])OP(=O)([O-])[O-])C"
                + "(O)C2O)c(=O)[nH]1.O>>Nc1nc2c(ncn2C2OC(COP(=O)([O-])[O-])C"
                + "(O)C2O)c(=O)[nH]1"
            ),
            ["O=[PH]([O-])OP(=O)([O-])[O-]", None],
            [[{"P": 1}], None],
            [[{"O": 16}], None],
            "",
            [
                (
                    "Nc1nc2c(ncn2C2OC(COP(=O)([O-])OP(=O)([O-])OP(=O)([O-])"
                    + "[O-])C(O)C2O)c(=O)[nH]1"
                ),
                "O",
            ],
            [
                (
                    "[#7]-&!@[#6]1:&@[#7]:&@[#6]2:&@[#6](:&@[#6](:&@[#7]:&@1)"
                    + "=&!@[#8]):&@[#7]:&@[#6]:&@[#7]:&@2-&!@[#6]1-&@[#8]-&@"
                    + "[#6](-&@[#6](-&@[#6]-&@1-&!@[#8])-&!@[#8])-&!@[#6]-&!"
                    + "@[#8]-&!@[#15](=&!@[#8])(-&!@[#8])-&!@[#8]"
                ),
                "",
            ],
        ),
    ],
)
def test_two_boundaries(rxn, exp_smiles, exp_ba, exp_nn, exp_issue, exp_sr, exp_mcs):
    mcs = _init_mcs()
    data = []
    _add(data, rxn)

    results = mcs.find(data)

    result = results[0][MCS_DATA_COL]
    assert exp_smiles == result["smiles"]
    assert exp_ba == result["boundary_atoms_products"]
    assert exp_nn == result["nearest_neighbor_products"]
    assert exp_issue == result["issue"]
    assert exp_sr == result["sorted_reactants"]
    assert exp_mcs == result["mcs_results"]



import unittest

from synrbl.SynRuleImputer import AutomaticRulesExtraction


class TestAutomaticRulesExtraction(unittest.TestCase):
    def setUp(self):
        self.extractor = AutomaticRulesExtraction()

    def test_smiles_to_molecular_formula(self):
        # Test conversion from SMILES to molecular formula
        smiles = "CCO"
        formula = AutomaticRulesExtraction.smiles_to_molecular_formula(smiles)
        self.assertEqual(formula, "C2H6O")

    def test_add_new_entries(self):
        # Test adding new entries
        filtered_fragments = {"smiles": ["CCO", "CC"]}
        self.extractor.add_new_entries(filtered_fragments)
        self.assertEqual(len(self.extractor.new_smiles_dict), 2)
        self.assertEqual(self.extractor.new_smiles_dict[0]["smiles"], "CCO")

    def test_extract_rules(self):
        # Test extracting rules
        filtered_fragments = {"smiles": ["CCO", "CC"]}
        self.extractor.add_new_entries(filtered_fragments)
        rules = self.extractor.extract_rules()
        self.assertEqual(len(rules), 2)
        self.assertIn("smiles", rules[0])
        self.assertIn("formula", rules[0])


if __name__ == "__main__":
    unittest.main()



import pytest

from synrbl import Balancer


def test_e2e_1():
    n = 100
    reactant = "[Br]" + n * "[Si](C)(C)O" + "[Si](C)(C)[Br]"
    product = "O" + n * "[Si](C)(C)O" + "[Si](C)(C)O"
    reaction = "{}>>{}".format(reactant, product)
    exp_result = "{}.{}>>{}.{}".format(reactant, "O.O", product, "Br.Br")

    balancer = Balancer()
    balancer.rb_method.n_jobs = 1
    balancer.rb_validator.n_jobs = 1
    balancer.mcs_validator.n_jobs = 1
    balancer.input_validator.n_jobs = 1

    result = balancer.rebalance(reaction)
    assert exp_result == result[0]


@pytest.mark.parametrize(
    "smiles,exp_smiles",
    [
        ["CC(=O)C>>CC(O)C", "CC(=O)C.[HH]>>CC(O)C"],
        [
            "CCO.[O]>>CC=O",
            "CCO.O=[Cr](Cl)(-[O-])=O.c1cc[nH+]cc1.O>>"
            + "CC=O.O.O=[Cr](O)O.c1cc[nH+]cc1.[Cl-]",
        ],
    ],
)
def test_post_process(smiles, exp_smiles):
    blncer = Balancer(n_jobs=1)
    result = blncer.rebalance(smiles)
    assert exp_smiles == result[0]


def test_standardizing_of_merged_compounds():
    blncer = Balancer(n_jobs=1)
    smiles = "CC(=O)OC=C>>CC(=O)O"
    exp_result = "CC(=O)OC=C.O>>CC(=O)O.CC=O"
    result = blncer.rebalance(smiles)
    assert exp_result == result[0]



from synrbl.SynProcessor import RSMIComparator, RSMIDecomposer, CheckCarbonBalance
from synrbl.SynUtils.common import update_reactants_and_products


class Validator:
    def __init__(
        self,
        reaction_col,
        method,
        solved_col="solved",
        solved_method_col="solved_by",
        unbalance_col="unbalance",
        carbon_balance_col="carbon_balance_check",
        issue_col="issue",
        check_carbon_balance=True,
        n_jobs=1,
    ):
        self.reaction_col = reaction_col
        self.method = method
        self.solved_col = solved_col
        self.solved_method_col = solved_method_col
        self.unbalance_col = unbalance_col
        self.check_carbon_balance = check_carbon_balance
        self.carbon_balance_col = carbon_balance_col
        self.issue_col = issue_col
        self.n_jobs = n_jobs

    def check(self, reactions, override_unsolved=False):
        update_reactants_and_products(reactions, self.reaction_col)
        decompose = RSMIDecomposer(
            smiles=None,  # type: ignore
            data=reactions,  # type: ignore
            reactant_col="reactants",
            product_col="products",
            parallel=True,
            n_jobs=self.n_jobs,
            verbose=0,
        )
        react_dict, product_dict = decompose.data_decomposer()

        comp = RSMIComparator(
            reactants=react_dict,  # type: ignore
            products=product_dict,  # type: ignore
            n_jobs=self.n_jobs,
            verbose=0,
        )
        unbalance, _ = comp.run_parallel(reactants=react_dict, products=product_dict)

        if self.check_carbon_balance:
            check = CheckCarbonBalance(
                reactions,
                rsmi_col=self.reaction_col,
                symbol=">>",
                atom_type="C",
                n_jobs=self.n_jobs,
            )
            for i, r in enumerate(check.check_carbon_balance()):
                reactions[i][self.carbon_balance_col] = r["carbon_balance_check"]

        assert len(reactions) == len(unbalance)
        for reaction, b in zip(reactions, unbalance):
            reaction[self.unbalance_col] = b
            if (
                b == "Balance"
                and reaction[self.carbon_balance_col] == "balanced"
                and not reaction[self.solved_col]
            ):
                reaction[self.solved_col] = True
                reaction[self.solved_method_col] = self.method
            if override_unsolved and not reaction[self.solved_col]:
                reaction[self.reaction_col] = reaction["input_reaction"]
        return reactions



from .data_utils import (
    save_database,
    load_database,
    extract_atomic_elements,
    find_shortest_sublists,
    filter_data,
    remove_duplicates_by_key,
    sort_by_key_length,
    add_missing_key_to_dicts,
    extract_results_by_key,
    get_random_samples_by_key,
)
from .chem_utils import (
    CheckCarbonBalance,
    calculate_net_charge,
    remove_atom_mapping,
    normalize_smiles,
    wc_similarity,
)
from .batching import Dataset, DataLoader



import unittest
from synrbl.SynChemImputer.curate_oxidation import CurationOxidation


class TestCurationOxidation(unittest.TestCase):

    def setUp(self):
        self.curate = CurationOxidation()
        self.data = [
            {"R-id": "alcohol_carbonyl", "reactions": "CCO.[O]>>CC=O"},
            {"R-id": "alcohol_acid", "reactions": "CCO.[O]>>CC(=O)O"},
            {"R-id": "carbonyl_acid", "reactions": "CC=O.[O]>>CC(=O)O"},
        ]

    def test_alcohol_carbonyl(self):
        result = self.curate.process_dict(
            self.data[0],
            "reactions",
            return_all=True,
            compounds_template=self.curate.compounds_template,
            reaction_templates=self.curate.reaction_templates,
        )
        self.assertEqual(
            result["curated_reaction"][0],
            "CCO.O=[Cr](Cl)(-[O-])=O.c1cc[nH+]cc1>>CC=O.O=[Cr](O)O.c1cc[nH+]cc1.[Cl-]",
        )
        self.assertEqual(result["stoichiometric"][0], [1])

    def test_alcohol_acid(self):
        result = self.curate.process_dict(
            self.data[1],
            "reactions",
            return_all=True,
            compounds_template=self.curate.compounds_template,
            reaction_templates=self.curate.reaction_templates,
        )
        print(result["curated_reaction"][0])
        self.assertEqual(
            result["curated_reaction"][0],
            (
                "CCO.[K][O][Mn](=O)(=O)=O.OS(=O)(=O)O>>"
                + "CC(=O)O.[K][O]S(=O)(=O)[O][K].[Mn]1[O]S(=O)(=O)[O]1"
            ),
        )
        self.assertEqual(result["stoichiometric"][0], [5, 4, 6, 5, 11, 2, 4])

    def test_aldehyde_acid(self):
        result = self.curate.process_dict(
            self.data[2],
            "reactions",
            return_all=True,
            compounds_template=self.curate.compounds_template,
            reaction_templates=self.curate.reaction_templates,
        )
        self.assertEqual(
            result["curated_reaction"][0],
            "CC=O.O=[Mn](=O)(=O)O[K].O>>CC(=O)O.O=[Mn]=O.O[K]",
        )
        self.assertEqual(result["stoichiometric"][0], [5, 4, 6, 5, 4, 1])


if __name__ == "__main__":
    unittest.main()



import json
import argparse
import logging
import pandas as pd
import rdkit.Chem.rdChemReactions as rdChemReactions

from argparse import RawTextHelpFormatter

from synrbl import Balancer

logger = logging.getLogger(__name__)


def print_result(stats, min_confidence=0):
    def _sr(v, c):
        return "{:.2%}".format(v / c) if c > 0 else "-"

    rxn_cnt = stats["reaction_cnt"]
    in_blcd = stats["balanced_cnt"]
    rb_s = stats["rb_solved"]
    rb_a = stats["rb_applied"]
    mcs_s = stats["mcs_solved"]
    mcs_a = stats["mcs_applied"]
    mcs_cth = stats["confident_cnt"]
    logger.info("{} Summary {}".format("#" * 20, "#" * 20))
    logger.info(
        "Input data contained {} balanced reactions.".format(stats["balanced_cnt"])
    )
    logger.info(
        "Rule-based method solved {} out of {} reactions (success rate: {}).".format(
            rb_s, rb_a, _sr(rb_s, rb_a)
        )
    )
    logger.info(
        "MCS-based method solved {} out of {} reactions (success rate: {}).".format(
            mcs_s, mcs_a, _sr(mcs_s, mcs_a)
        )
    )
    below_th = mcs_s - mcs_cth
    if below_th > 0:
        logger.info(
            "{} results where below the confidence threshold of {:.0%}.".format(
                below_th, min_confidence
            )
        )
        logger.info(
            (
                "MCS-based method solved {} out of {} reactions above the "
                + "confidence threshold (success rate: {})."
            ).format(mcs_cth, mcs_a, _sr(mcs_cth, mcs_a))
        )
    logger.info(
        "SynRBL solved {} out of {} reactions (success rate: {}).".format(
            rb_s + mcs_cth, rxn_cnt - in_blcd, _sr(rb_s + mcs_cth, rxn_cnt - in_blcd)
        )
    )


def check_columns(reactions, reaction_col, required_cols=[]):
    if len(reactions) == 0:
        raise ValueError("No reactions found in input.")
    cols = reactions[0].keys()
    if reaction_col not in cols:
        raise KeyError("No column '{}' found in input.".format(reaction_col))
    if not isinstance(reactions[0][reaction_col], str):
        raise TypeError(
            "Reaction column '{}' must be of type string not '{}'.".format(
                reaction_col, type(reactions[0][reaction_col])
            )
        )

    mol = None
    try:
        mol = rdChemReactions.ReactionFromSmarts(
            reactions[0][reaction_col], useSmiles=True
        )
    except Exception:
        pass
    if mol is None:
        raise ValueError(
            "Value '{}...' in reaction column '{}' is not a valid SMILES.".format(
                reactions[0][reaction_col][0:30], reaction_col
            )
        )
    for c in required_cols:
        if c not in reactions[0].keys():
            raise KeyError(
                (
                    "Required column '{}' not found. The input to benchamrk "
                    + "should be the output from a rebalancing run."
                ).format(c)
            )


def impute(
    src_file,
    output_file,
    reaction_col,
    passthrough_cols,
    min_confidence,
    n_jobs=-1,
    cache=False,
    cache_dir=None,
    batch_size=None,
):
    input_reactions = pd.read_csv(src_file).to_dict("records")
    check_columns(input_reactions, reaction_col, required_cols=passthrough_cols)

    synrbl = Balancer(
        reaction_col=reaction_col,
        confidence_threshold=min_confidence,
        n_jobs=n_jobs,
        cache=cache,
        cache_dir=cache_dir,
        batch_size=batch_size,
    )
    stats = {}
    rbl_reactions = synrbl.rebalance(input_reactions, output_dict=True, stats=stats)

    for in_r, out_r in zip(input_reactions, rbl_reactions):
        for c in passthrough_cols:
            out_r[c] = in_r[c]

    df = pd.DataFrame(rbl_reactions)
    df.to_csv(output_file)
    with open("{}.stats".format(output_file), "w") as f:
        json.dump(stats, f, indent=4)

    print_result(stats, min_confidence)


def run(args):
    outputfile = args.o
    if outputfile is None:
        outputfile = "{}_out.csv".format(args.inputfile.split(".")[0])
    passthrough_cols = (
        args.out_columns if isinstance(args.out_columns, list) else [args.out_columns]
    )
    batch_size = None
    if len(args.batch_size) > 0:
        batch_size = int(args.batch_size)

    impute(
        args.inputfile,
        outputfile,
        reaction_col=args.col,
        passthrough_cols=passthrough_cols,
        min_confidence=args.min_confidence,
        n_jobs=args.p,
        cache=args.cache,
        cache_dir=args.cache_dir,
        batch_size=batch_size,
    )


class Range(object):
    def __init__(self, start, end):
        self.start = start
        self.end = end

    def __eq__(self, other):
        return self.start <= other <= self.end

    def __str__(self):
        return "[{}, {}]".format(self.start, self.end)


def list_of_strings(arg):
    return arg.split(",")


def configure_argparser(argparser: argparse._SubParsersAction):
    default_cache_dir = "./cache"
    default_min_confidence = 0
    default_num_processes = -1
    default_reaction_col = "reaction"

    run_parser = argparser.add_parser(
        "run",
        description="Try to rebalance chemical reactions.",
        formatter_class=RawTextHelpFormatter,
    )

    run_parser.add_argument(
        "inputfile",
        help=(
            "Path to file containing reaction SMILES. \n"
            + "Possible file formats are:\n"
            + "  - csv: SMILES are expected to be in column '{}'\n".format(
                default_reaction_col
            )
            + "         The column name can be changed with the --col\n"
            + "         argument.\n"
            + "  - json: The expected format is a list of objects:\n"
            + "          [{...},{...}]\n"
            + "          The reaction SMILES is expected as value of\n"
            + "          '{}'. The reaction key can be changed with\n".format(
                default_reaction_col
            )
            + "          the --col argument."
        ),
    )
    run_parser.add_argument("-o", default=None, help="Path to output file.")
    run_parser.add_argument(
        "-p",
        default=default_num_processes,
        type=int,
        help=("The number of parallel process. (Default {} => # of processors)").format(
            default_num_processes
        ),
    )
    run_parser.add_argument(
        "-b",
        "--batch-size",
        default=None,
        help=("Number of reactions that should be processed at once."),
    )
    run_parser.add_argument(
        "--col",
        default=default_reaction_col,
        help=(
            "The reaction SMILES column name (csv) or key (json) in the \n"
            + "input file. (Default: '{}')"
        ).format(default_reaction_col),
    )
    run_parser.add_argument(
        "--out-columns",
        default=[],
        type=list_of_strings,
        help="A comma separated list of columns/keys from the input \n"
        + "that should be added to the output. (e.g.: col1,col2,col3)",
    )
    run_parser.add_argument(
        "--min-confidence",
        type=float,
        default=default_min_confidence,
        choices=[Range(0.0, 1.0)],
        help=(
            "Set a confidence threshold for the results from the\n"
            + "MCS-based method. (Default: {})"
        ).format(default_min_confidence),
    )
    run_parser.add_argument(
        "--cache",
        action="store_true",
        help=(
            "Flag to cache intermediate results. Use this together \n"
            + "with --batch-size."
        ),
    )
    run_parser.add_argument(
        "--cache-dir",
        default=default_cache_dir,
        help=(
            "The directory that is used to store intermediate results. \n"
            + "(Default: {})".format(default_cache_dir)
        ),
    )

    run_parser.set_defaults(func=run)



from typing import List, Dict


class GraphMissingUncertainty:
    """
    A class to handle uncertainty in graph data. It identifies uncertain
    elements in the graphs based on boundary conditions and fragment count.

    Attributes:
        missing_graph_list (List[Dict]): A list of graph data in dictionary
            format.
        threshold (int): A threshold value to determine uncertainty based on
            the number of fragments in a SMILES string.
    """

    def __init__(self, missing_graph_list: List[Dict], threshold: int = 2) -> None:
        """
        Initializes the GraphMissingUncertainty class with missing graph data
        and a threshold.

        Args:
            missing_graph_list (List[Dict]): The list of missing graph data.
            threshold (int, optional): The threshold for determining
                uncertainty based on fragments. Defaults to 2.
        """
        self.missing_graph_list = missing_graph_list
        self.threshold = threshold

    @staticmethod
    def check_boundary(data_list: List[Dict]) -> List[int]:
        """
        Identifies the indices of entries in the data list where
        'boundary_atoms_products' is empty.

        Args:
            data_list (List[Dict]): A list of dictionaries, each containing a
                key 'boundary_atoms_products'.

        Returns:
            List[int]: A list of indices where 'boundary_atoms_products' is
                empty.
        """
        without_boundary_key = []

        for key, item in enumerate(data_list):
            if all(
                element is None for element in item["boundary_atoms_products"]
            ):  # Checks if 'boundary_atoms_products' is empty
                without_boundary_key.append(key)

        return without_boundary_key

    @staticmethod
    def check_fragments(data_list: List[Dict], threshold: int = 2) -> List[int]:
        """
        Identifies the indices of entries in the data list where the number of
        SMILES string fragments meets or exceeds a specified threshold.

        Args:
            data_list (List[Dict]): A list of dictionaries, each containing a
                'smiles' key.
            threshold (int): The minimum number of fragments in a SMILES string
                to be considered uncertain.

        Returns:
            List[int]: A list of indices where the number of SMILES string
                fragments meets or exceeds the threshold.
        """
        graph_uncertain_key = []

        for key, entry in enumerate(data_list):
            for i in entry["smiles"]:
                if i is not None and len(i.split(".")) >= threshold:
                    graph_uncertain_key.append(key)

        return graph_uncertain_key

    def fit(self) -> List[Dict]:
        """
        Processes the missing graph data to update their 'Certainty' status
        based on boundary and fragment checks.

        Returns:
            List[Dict]: The updated list of missing graph data with 'Certainty'
                status marked.
        """
        uncertain_key = []
        without_boundary_key = self.check_boundary(self.missing_graph_list)
        graph_uncertain_key = self.check_fragments(
            self.missing_graph_list, threshold=self.threshold
        )

        uncertain_key.extend(without_boundary_key)
        uncertain_key.extend(graph_uncertain_key)

        for key, missing_graph in enumerate(self.missing_graph_list):
            missing_graph["Certainty"] = key not in uncertain_key

        return self.missing_graph_list



import io

from rdkit import Chem
from rdkit.Chem.Draw import rdMolDraw2D
from PIL import Image


class MCSVisualizer:
    """
    A class for visualizing molecules and highlighting their Maximum Common
    Substructure (MCS).

    Methods
    -------
    draw_molecule_with_atom_numbers(mol):
        Annotates a molecule with atom indices.

    def highlight_molecule(
        molecule_smiles,
        mcs_smiles,
        show_atom_numbers=False,
        compare=False,
        missing_graph_smiles=None,
    ):
        Highlights the specified substructure within a molecule and optionally
        compares it with another molecule.
    """

    def __init__(self):
        pass

    def draw_molecule_with_atom_numbers(self, mol):
        """
        Annotates a molecule with atom indices.

        Parameters
        ----------
        mol : RDKit Molecule object
            The molecule to be annotated with atom numbers.

        Returns
        -------
        RDKit Mol object
            The annotated molecule.
        """
        mol_with_atom_numbers = Chem.Mol(mol)
        for atom in mol_with_atom_numbers.GetAtoms():
            atom.SetProp("atomLabel", str(atom.GetIdx()))
        return mol_with_atom_numbers

    def highlight_molecule(
        self,
        molecule_smiles,
        mcs_smiles,
        show_atom_numbers=False,
        compare=False,
        missing_graph_smiles=None,
    ):
        """
        Highlights the specified substructure (MCS) within a molecule and
        optionally compares it with another molecule.

        Parameters
        ----------
        molecule_smiles : str
            SMILES representation of the molecule.
        mcs_smiles : str
            SMILES representation of the substructure (MCS).
        show_atom_numbers : bool, optional
            If True, shows atom numbers on the molecules.
        compare : bool, optional
            If True, shows a comparison with another molecule specified by
            missing_graph_smiles.
        missing_graph_smiles : str, optional
            SMILES representation of the second molecule for comparison.

        Returns
        -------
        PIL.Image.Image
            An image of the molecule with the specified substructure
            highlighted, and optionally the second molecule.

        Example
        -------
        visualizer = MCSVisualizer()
        img = visualizer.highlight_molecule(
            "CCO", "CO", compare=True, missing_graph_smiles="CCC"
        )
        img.show()
        """
        molecule = Chem.MolFromSmiles(molecule_smiles)
        pattern = Chem.MolFromSmarts(Chem.CanonSmiles(mcs_smiles))
        matching = molecule.GetSubstructMatch(pattern)

        if not matching:
            raise ValueError("Substructure not found in molecule.")

        if show_atom_numbers:
            molecule = self.draw_molecule_with_atom_numbers(molecule)

        mols = [molecule]
        highlight_lists = [matching] if matching else [None]
        legends = ["Molecule with MCS Highlighted"]

        # Check if a comparison molecule is needed
        if compare and missing_graph_smiles:
            missing_graph = Chem.MolFromSmiles(missing_graph_smiles)
            if show_atom_numbers:
                missing_graph = self.draw_molecule_with_atom_numbers(missing_graph)
            mols.append(missing_graph)
            highlight_lists.append(None)
            legends.append("Missing Molecule")

        # Image dimensions
        width, height = 600, 600
        molsPerRow = 2 if compare else 1
        img_width = width * molsPerRow
        img_height = height * len(mols) // molsPerRow

        # Create a drawing object
        drawer = rdMolDraw2D.MolDraw2DCairo(img_width, img_height, width, height)

        # Draw each molecule
        for i, mol in enumerate(mols):
            row = i // molsPerRow
            col = i % molsPerRow
            x_offset = width * col
            y_offset = height * row
            drawer.SetOffset(x_offset, y_offset)
            drawer.DrawMolecule(
                mol, legend=legends[i], highlightAtoms=highlight_lists[i]
            )

        drawer.FinishDrawing()

        return Image.open(io.BytesIO(drawer.GetDrawingText()))



import unittest
import rdkit.Chem.rdmolfiles as rdmolfiles
import synrbl.SynUtils.functional_group_utils as fgutils


class TestReduce(unittest.TestCase):
    def test_1(self):
        mol = rdmolfiles.MolFromSmiles("CCCOC(C)=O")
        mol_r, _ = fgutils.trim_mol(mol, 6, 1)
        self.assertEqual("C=O", rdmolfiles.MolToSmiles(mol_r))

    def test_2(self):
        mol = rdmolfiles.MolFromSmiles("CCCOC(C)=O")
        mol_r, _ = fgutils.trim_mol(mol, 3, 1)
        self.assertEqual("COC", rdmolfiles.MolToSmiles(mol_r))

    def test_3(self):
        mol = rdmolfiles.MolFromSmiles("CCCOC(C)=O")
        mol_r, _ = fgutils.trim_mol(mol, 4, 1)
        self.assertEqual("CC(=O)O", rdmolfiles.MolToSmiles(mol_r))

    def test_4(self):
        mol = rdmolfiles.MolFromSmiles("COc1ccccc1")
        mol_r, _ = fgutils.trim_mol(mol, 1, 1)
        self.assertEqual("COc", rdmolfiles.MolToSmiles(mol_r))


class TestMappingPermutations(unittest.TestCase):
    def test_1(self):
        self.assertEqual([[]], fgutils.get_mapping_permutations([], ["C"]))

    def test_2(self):
        self.assertEqual(
            [[(0, 1)]], fgutils.get_mapping_permutations(["C"], ["S", "C"])
        )

    def test_3(self):
        self.assertEqual(
            [[(0, 0)], [(0, 1)]], fgutils.get_mapping_permutations(["C"], ["C", "C"])
        )

    def test_4(self):
        self.assertEqual(
            [[(0, 0), (1, 1)], [(0, 1), (1, 0)]],
            fgutils.get_mapping_permutations(["C", "C"], ["C", "C"]),
        )


class TestMCS(unittest.TestCase):
    def _call(self, smiles, anchor, pattern_smiles, panchor=None, expected_match=True):
        mol = rdmolfiles.MolFromSmiles(smiles)
        pattern_mol = rdmolfiles.MolFromSmiles(pattern_smiles)
        match, mapping = fgutils.pattern_match(
            mol, anchor, pattern_mol, pattern_anchor=panchor
        )
        self.assertEqual(expected_match, match)
        return mapping

    def test_simple_pattern_match(self):
        self.assertEqual([(2, 0)], self._call("CCO", 2, "O", 0))
        self.assertEqual([(2, 0), (1, 1)], self._call("CCO", 2, "OC", 0))
        self.assertEqual([(1, 1), (2, 0)], self._call("CCO", 1, "OC", 1))
        self.assertEqual([(2, 0), (1, 1), (0, 2)], self._call("CCO", 2, "OCC", 0))
        self.assertEqual([(2, 2), (1, 1), (0, 0)], self._call("CCO", 2, "CCO", 2))

    def test_ring_pattern_match(self):
        self.assertEqual([(1, 0), (3, 1), (2, 2)], self._call("CC1NO1", 1, "CON", 0))
        self.assertEqual([(1, 0), (3, 2), (2, 1)], self._call("CC1NO1", 1, "C1NO1", 0))

    def test_invalid_path_pattern(self):
        self.assertEqual(
            [(3, 0), (4, 1), (5, 2)], self._call("CC(O)CC(N)O", 3, "CCN", 0)
        )

    def test_match_star(self):
        self.assertEqual([(1, 1), (2, 2), (0, 0)], self._call("CC(C)C", 1, "CCC", 1))

    def test_find_no_pattern_match(self):
        self._call("CCO", 0, "O", 0, expected_match=False)
        self._call("CCO", 1, "O", 0, expected_match=False)
        self._call("CC1NO1", 0, "CON", 0, expected_match=False)
        self._call("CCNO", 1, "C1NO1", 0, expected_match=False)
        self._call("CC(O)CC(N)O", 2, "CCN", 0, expected_match=False)

    def test_fit_without_pattern_anchor(self):
        self.assertEqual([(0, 0), (1, 1), (2, 2)], self._call("CCO", 0, "CCO"))
        self.assertEqual([(1, 1), (2, 2), (0, 0)], self._call("CCO", 1, "CCO"))
        self.assertEqual([(2, 2), (1, 1), (0, 0)], self._call("CCO", 2, "CCO"))

    def test_match_with_bond_order(self):
        self._call("CC=O", 1, "CO", expected_match=False)
        self.assertEqual(
            [(1, 0), (2, 1)], self._call("CC=O", 1, "C=O", expected_match=True)
        )

    def test_find_double_bond_match(self):
        self.assertEqual(
            [(1, 1), (2, 0)], self._call("CP(=O)(O)O", 1, "O=P", expected_match=True)
        )


class TestFGConfig(unittest.TestCase):
    def test_init_anti_pattern(self):
        config = fgutils.FGConfig("CN", anti_pattern=["C=O", "Nc1ccccc1", "CCC"])
        self.assertEqual(3, len(config.anti_pattern))
        self.assertEqual("Nc1ccccc1", rdmolfiles.MolToSmiles(config.anti_pattern[0]))
        self.assertEqual("CCC", rdmolfiles.MolToSmiles(config.anti_pattern[1]))
        self.assertEqual("C=O", rdmolfiles.MolToSmiles(config.anti_pattern[2]))

    def test_init_with_multiple_patterns(self):
        config = fgutils.FGConfig(["O=CS", "OC=S"])
        self.assertEqual(2, len(config.pattern))
        self.assertEqual(2, len(config.groups))
        self.assertEqual("O=CS", rdmolfiles.MolToSmiles(config.groups[0]))
        self.assertEqual("OC=S", rdmolfiles.MolToSmiles(config.groups[1]))


class TestFunctionalGroupCheck(unittest.TestCase):
    def __test_fg(self, smiles, group_name, indices=None):
        groups = list(fgutils.functional_group_config.keys())
        assert group_name in groups, "Functional group {} is not implemented.".format(
            group_name
        )

        mol = rdmolfiles.MolFromSmiles(smiles)
        indices = (
            [i for i in range(len(mol.GetAtoms()))] if indices is None else indices
        )
        for a in mol.GetAtoms():
            idx = a.GetIdx()
            pos_groups = []
            for g in groups:
                result = fgutils.is_functional_group(mol, g, idx)
                if result:
                    pos_groups.append(g)
            if idx in indices:
                if len(pos_groups) == 1 and group_name in pos_groups:
                    pass
                else:
                    self.assertTrue(
                        False,
                        (
                            "Atom {} ({}) in {} was expected to be in functional "
                            + "group {} but was identified as {}."
                        ).format(a.GetSymbol(), idx, smiles, group_name, pos_groups),
                    )
            else:
                self.assertTrue(
                    group_name not in pos_groups,
                    (
                        "Atom {} ({}) in {} was expected to be not part of "
                        + "functional group {} but was identified as {}."
                    ).format(a.GetSymbol(), idx, smiles, group_name, pos_groups),
                )

    def test_phenol(self):
        # 2,3 Xylenol
        self.__test_fg("Cc1cccc(O)c1C", "phenol", [1, 2, 3, 4, 5, 6, 7])
        self.__test_fg("Oc1ccc[nH]1", "phenol")

    def test_alcohol(self):
        # Ethanol
        self.__test_fg("CCO", "alcohol", [1, 2])

    def test_ether(self):
        # Methylether
        self.__test_fg("COC", "ether", [0, 1, 2])
        self.__test_fg("COc1cccc(OC)c1OC", "ether", [0, 1, 2, 9, 10, 11, 6, 7, 8])
        self.__test_fg("COc1ccc[nH]1", "ether", [0, 1, 2])
        self.__test_fg("COc1ccccc1", "ether", [0, 1, 2])

    def test_enol(self):
        # 3-pentanone enol
        self.__test_fg("CCC(O)=CC", "enol", [2, 3, 4])

    def test_amid(self):
        # Asparagine
        self.__test_fg("NC(=O)CC(N)C(=O)O", "amid", [0, 1, 2])

    # def test_acyl(self):
    #    # Acetyl cloride
    #    self.__test_fg("CC(=O)[Cl]", "acyl", [1, 2])

    def test_diol(self):
        self.__test_fg("OCO", "diol")

    def test_hemiacetal(self):
        self.__test_fg("COCO", "hemiacetal")

    def test_acetal(self):
        self.__test_fg("COCOC", "acetal")

    def test_urea(self):
        self.__test_fg("O=C(N)O", "urea")

    def test_carbonat(self):
        self.__test_fg("O=C(O)O", "carbonat")
        self.__test_fg("COC(=O)O", "carbonat", [1, 2, 3, 4])

    def test_ester(self):
        # Methyl acetate
        self.__test_fg("COC(C)=O", "ester", [1, 2, 4])

    def test_anhydrid(self):
        self.__test_fg("O=C(C)OC=O", "anhydrid")

    def test_acid(self):
        # Acetic acid
        self.__test_fg("CC(=O)O", "acid", [1, 2, 3])

    def test_anilin(self):
        self.__test_fg("Nc1ccccc1", "anilin")

    def test_amin(self):
        # Glycin
        self.__test_fg("NCC(=O)O", "amin", [0, 1])
        # Methcatione
        self.__test_fg("CNC(C)C(=O)c1ccccc1", "amin", [0, 1, 2])

    def test_nitril(self):
        self.__test_fg("C#N", "nitril")

    def test_hydroxylamin(self):
        self.__test_fg("NO", "hydroxylamin")

    def test_nitrose(self):
        self.__test_fg("N=O", "nitrose")

    def test_nitro(self):
        self.__test_fg("ON=O", "nitro")

    def test_thioether(self):
        # Diethylsulfid
        self.__test_fg("CCSCC", "thioether", [1, 2, 3])

    def test_thioester(self):
        # Methyl thionobenzonat
        self.__test_fg("CSC(=O)c1ccccc1", "thioester", [1, 2, 3])
        self.__test_fg("COC(=S)c1ccccc1", "thioester", [1, 2, 3])

    def test_keton(self):
        # Methcatione
        self.__test_fg("CNC(C)C(=O)c1ccccc1", "keton", [4, 5])

    def test_aldehyde(self):
        # Methcatione
        self.__test_fg("CC=O", "aldehyde", [1, 2])



from joblib import Parallel, delayed


class BothSideReact:
    """
    Class to process chemical reactions on both sides (reactants and products).

    Attributes:
    react_dict (list): A list of dictionaries representing reactants.
    product_dict (list): A list of dictionaries representing products.
    unbalance (list): A list indicating the balance status of each reaction.
    diff_formula (list): A list containing the differential formula for
        each reaction.
    """

    def __init__(self, react_dict, product_dict, unbalance, diff_formula):
        """
        Initializes the BothSideReact class with reaction dictionaries,
        unbalance, and differential formulas.

        Parameters:
        react_dict (list): List of dictionaries representing reactants.
        product_dict (list): List of dictionaries representing products.
        unbalance (list): List indicating the balance status of each reaction.
        diff_formula (list): List containing the differential formula for
            each reaction.
        """
        self.react_dict = react_dict
        self.product_dict = product_dict

        # Ensure 'Q' key is present in all reactant and product dictionaries
        for d in [self.react_dict, self.product_dict]:
            for value in d:
                if "Q" not in value:
                    value["Q"] = 0

        self.unbalance = unbalance
        self.diff_formula = diff_formula

    @staticmethod
    def enforce_product_side(react_dict, product_dict):
        """
        Enforces the product side of the reaction by calculating the difference
        between reactant and product counts.

        Parameters:
        react_dict (dict): Dictionary representing a single reactant.
        product_dict (dict): Dictionary representing a single product.

        Returns:
        dict: Dictionary representing the differential count between reactant
            and product.
        """
        diff_dict = {}
        # Calculate the difference between reactants and products
        for key, value in react_dict.items():
            diff_value = value - product_dict.get(key, 0)
            if diff_value != 0:
                diff_dict[key] = diff_value

        # Add keys only present in product_dict
        for key, value in product_dict.items():
            if key not in react_dict:
                diff_dict[key] = -value

        return diff_dict

    @staticmethod
    def filter_list_by_indices(data, indices):
        """
        Filters a list by specified indices.

        Parameters:
        data (list): The original list to filter.
        indices (list): List of indices to include in the filtered list.

        Returns:
        list: A filtered list containing elements at the specified indices.
        """
        # Use list comprehension for efficient filtering
        return [data[i] for i in indices if i < len(data)]

    @staticmethod
    def reverse_values_if_negative_except_Q(diff_dict):
        """
        Reverses the values in the dictionary if negative, except for the key 'Q'.

        Parameters:
        diff_dict (dict): The dictionary with potential negative values.

        Returns:
        tuple: A tuple containing the updated dictionary and a string
            indicating the balance status.
        """
        if len(diff_dict) == 2 and "Q" in diff_dict.keys():
            if any(value < 0 for key, value in diff_dict.items() if key != "Q"):
                # Reverse all values except for 'Q'
                return {key: -value for key, value in diff_dict.items()}, "Reactants"
            else:
                # Original dictionary if no negative values found
                return diff_dict, "Products"
        else:
            # Return as is if conditions not met
            return diff_dict, "Both"

    def fit(self, n_jobs=4):
        """
        Processes the reactions by balancing reactants and products and
        updating the unbalance status.

        Returns:
        tuple: A tuple containing the updated diff_formula and unbalance lists.
        """
        # Filter indices where balance status is 'Both'
        both_index = [i for i, val in enumerate(self.unbalance) if val == "Both"]
        react_dict_both = self.filter_list_by_indices(self.react_dict, both_index)
        product_dict_both = self.filter_list_by_indices(self.product_dict, both_index)

        # Process reactions in parallel for efficiency
        diff_dict = Parallel(n_jobs=n_jobs, verbose=0)(
            delayed(self.enforce_product_side)(react, prod)
            for react, prod in zip(react_dict_both, product_dict_both)
        )

        # Post-process to determine the new balance status
        diff_dict_both, unbalance_both = [], []
        for item in diff_dict:
            d, u = self.reverse_values_if_negative_except_Q(item)
            diff_dict_both.append(d)
            unbalance_both.append(u)

        # Update diff_formula and unbalance lists
        for index, diff_new, unbalance_new in zip(
            both_index, diff_dict_both, unbalance_both
        ):
            self.diff_formula[index] = diff_new
            self.unbalance[index] = unbalance_new

        return self.diff_formula, self.unbalance



import unittest

from synrbl.SynRuleImputer import RuleImputeManager


class TestRuleImputeManager(unittest.TestCase):
    def setUp(self):
        self.db = RuleImputeManager()

    def test_add_entry_valid(self):
        # Test adding a valid entry
        formula = "H2O"
        smiles = "O"
        self.db.add_entry(formula, smiles)
        self.assertEqual(len(self.db.database), 1)
        self.assertEqual(self.db.database[0]["formula"], formula)
        self.assertEqual(self.db.database[0]["smiles"], smiles)

    def test_add_entry_duplicate_formula(self):
        # Test adding an entry with a duplicate formula
        self.db.add_entry("H2O", "O")
        with self.assertRaises(ValueError):
            self.db.add_entry("H2O", "O")

    def test_add_entry_duplicate_smiles(self):
        # Test adding an entry with a duplicate SMILES string
        self.db.add_entry("H2O", "O")
        with self.assertRaises(ValueError):
            self.db.add_entry("Water", "O")

    def test_add_entry_invalid_smiles(self):
        # Test adding an entry with an invalid SMILES string
        with self.assertRaises(ValueError):
            self.db.add_entry("Invalid", "InvalidString")

    def test_add_entries_mixed_validity(self):
        # Test adding multiple entries with mixed validity
        entries = [
            {"formula": "CO2", "smiles": "C=O"},
            {"formula": "Invalid", "smiles": "InvalidString"},
        ]
        invalid_entries = self.db.add_entries(entries)
        self.assertEqual(len(invalid_entries), 1)
        self.assertEqual(invalid_entries[0]["formula"], "Invalid")

    def test_remove_entry(self):
        # Test removing an entry
        self.db.add_entry("H2O", "O")
        self.db.remove_entry("H2O")
        self.assertEqual(len(self.db.database), 0)

    def test_canonicalize_smiles(self):
        # Test canonicalizing a SMILES string
        canonical_smiles = RuleImputeManager.canonicalize_smiles("C1=CC=CC=C1")
        self.assertEqual(canonical_smiles, "c1ccccc1")

    def test_is_valid_smiles(self):
        # Test validating a SMILES string
        self.assertTrue(RuleImputeManager.is_valid_smiles("C1=CC=CC=C1"))
        self.assertFalse(RuleImputeManager.is_valid_smiles("InvalidString"))


if __name__ == "__main__":
    unittest.main()



import unittest

from synrbl.SynRuleImputer import SyntheticRuleMatcher


class TestSyntheticRuleMatcher(unittest.TestCase):
    def setUp(self):
        # Sample rules and data
        self.rule_dict = [
            {"smiles": "C1=CC=CC=C1", "Composition": {"C": 6, "H": 6}},
            {"smiles": "CCO", "Composition": {"C": 2, "H": 6, "O": 1}},
        ]
        self.data_dict = {"C": 6, "H": 6}
        self.matcher = SyntheticRuleMatcher(
            self.rule_dict, self.data_dict, select="best", ranking=False
        )

    def test_match(self):
        # Test matching solutions
        solution = self.matcher.match()
        self.assertIsInstance(solution, list)
        self.assertGreater(len(solution), 0)

    def test_dfs(self):
        # Test depth-first search algorithm
        solution = self.matcher.dfs(self.data_dict, [])
        self.assertIsInstance(solution, list)
        self.assertGreater(len(solution), 0)

    def test_apply_rule(self):
        # Test applying a chemical rule
        rule = {"smiles": "C1=CC=CC=C1", "Composition": {"C": 6, "H": 6}}
        new_data, new_path = self.matcher.apply_rule(self.data_dict, [], rule)
        self.assertIsNotNone(new_data)
        self.assertIsNotNone(new_path)

    def test_can_match(self):
        # Test checking if a chemical rule can be matched
        rule = {"C": 6, "H": 6}
        can_match = self.matcher.can_match(rule, self.data_dict)
        self.assertTrue(can_match)

    def test_rank_solutions(self):
        # Test ranking solutions
        solutions = [[{"smiles": "C1=CC=CC=C1", "Ratio": 1}]]
        ranked_solutions = SyntheticRuleMatcher.rank_solutions(
            solutions, ranking="longest"
        )
        self.assertEqual(solutions, ranked_solutions)

    def test_remove_overlapping_solutions(self):
        # Test removing overlapping solutions
        solutions = [
            [{"smiles": "C1=CC=CC=C1", "Ratio": 1}],
            [{"smiles": "C1=CC=CC=C1", "Ratio": 1}],
        ]
        unique_solutions = SyntheticRuleMatcher.remove_overlapping_solutions(solutions)
        self.assertEqual(len(unique_solutions), 1)


if __name__ == "__main__":
    unittest.main()



import unittest

from rdkit import Chem

from synrbl.SynRuleImputer import SyntheticRuleImputer


# not fixed
class TestSyntheticRuleImputer(unittest.TestCase):
    def setUp(self):
        self.rule_dict = [
            {"smiles": "C1=CC=CC=C1", "Composition": {"C": 6, "H": 6}},
            {"smiles": "CCO", "Composition": {"C": 2, "H": 6, "O": 1}},
        ]
        self.imputer = SyntheticRuleImputer(self.rule_dict)

    def test_single_impute(self):
        # Test single imputation of missing data
        missing_data = {
            "Diff_formula": {"C": 4, "H": 4, "Q": 0},
            "Unbalance": "Products",
            "reactants": "C2H6",
            "products": "C2H6",
        }
        imputed_data = SyntheticRuleImputer.single_impute(missing_data, self.rule_dict)

        # Check if 'new_reaction' is added to the dictionary, which implies
        # successful imputation
        if any(rule["smiles"] in imputed_data["products"] for rule in self.rule_dict):
            self.assertIn("new_reaction", imputed_data)
        else:
            self.assertNotIn("new_reaction", imputed_data)

    def test_parallel_impute(self):
        # Test parallel imputation of missing data
        missing_data_list = [
            {
                "Diff_formula": {"C": 4, "H": 4},
                "Unbalance": "Products",
                "reactants": "C2H6",
                "products": "C2H6",
            },
            {
                "Diff_formula": {"C": 3, "H": 6},
                "Unbalance": "Reactants",
                "reactants": "C2H6",
                "products": "C2H6",
            },
        ]
        imputed_data_list = self.imputer.parallel_impute(missing_data_list)
        self.assertEqual(len(imputed_data_list), len(missing_data_list))

    def test_get_and_validate_smiles(self):
        # Test concatenation and validation of SMILES strings
        solution = [{"smiles": "CCO", "Ratio": 1}]
        smiles = SyntheticRuleImputer.get_and_validate_smiles(solution)
        self.assertIsInstance(smiles, str)
        self.assertTrue(Chem.MolFromSmiles(smiles) is not None)


if __name__ == "__main__":
    unittest.main()



import pandas as pd

from synrbl.SynProcessor import RSMIProcessing


def preprocess(reactions, reaction_col, index_col, solved_col, input_col, n_jobs=1):
    df = pd.DataFrame(reactions)
    df[solved_col] = False

    if "Unnamed: 0" in df.columns:
        df = df.drop("Unnamed: 0", axis=1)

    process = RSMIProcessing(
        data=df,
        rsmi_col=reaction_col,
        parallel=True,
        n_jobs=n_jobs,
        data_name=None,  # type: ignore
        index_col=index_col,
        drop_duplicates=False,
        save_json=False,
        save_path_name=None,  # type: ignore
        verbose=0,
    )
    reactions = process.data_splitter()
    reactions[input_col] = reactions[reaction_col]

    return reactions.to_dict("records")



from typing import List, Tuple, Optional
from collections import Counter


class RefinementUncertainty:
    def __init__(self, finalgraph_uncertainty, graph_conditions: List[dict]):
        """
        Initialize the RefinementUncertainty class with the graph conditions.

        Parameters:
        finalgraph_uncertainty (List[dict]): A list of dictionaries
            representing missing graph data.
        graph_conditions (List[dict]): A list of dictionaries representing
            graph conditions.
        """
        self.finalgraph_uncertainty = finalgraph_uncertainty
        self.id = [i["R-id"] for i in finalgraph_uncertainty]
        self.graph_conditions = graph_conditions

    @staticmethod
    def get_smiles_graph(final_graph: List[dict], id: str) -> List[str]:
        """
        Retrieve the SMILES graph for a given final graph and ID.

        Parameters:
        final_graph (List[dict]): A list of dictionaries representing missing
            graph data.
        id (str): The ID to search for in the graph.

        Returns:
        List[str]: The SMILES graph for the given ID.
        """
        return [item["smiles"] for item in final_graph if item["R-id"] == id][0]

    @staticmethod
    def intersection_of_lists_with_count(
        lists: List[List[str]], intersection_num: int = 2
    ) -> Tuple[Optional[List[str]], Optional[int]]:
        """
        Find lists that are present at least a specified number of times in
        the provided list of lists and their first index.

        Parameters:
        lists (List[List[str]]): A list containing multiple lists of elements.
        intersection_num (int): The minimum number of times a list must appear
            to be considered.

        Returns:
        Tuple[Optional[List[str]], Optional[int]]: The first list that meets
            the intersection criteria and its first index, or None if no such
            list exists.
        """
        # Convert inner lists to tuples for hashing and count occurrences
        tuple_lists = [tuple(lst) for lst in lists]
        counts = Counter(tuple_lists)

        # Find the first list that meets the intersection criteria
        for index, lst in enumerate(tuple_lists):
            if counts[lst] >= intersection_num:
                return list(lst), index  # Convert tuple back to list for the result

        return None, None

    def fit(self, intersection_num: int = 2) -> List[dict]:
        """
        Generate a new graph for uncertain IDs based on the first intersection
        condition.

        Parameters:
        id_uncertainty (List[str]): A list of uncertain IDs.
        intersection_num (int): The minimum number of lists in which an element
            must appear to be considered.

        Returns:
        List[dict]: A new graph list for uncertain IDs.
        """
        new_graph_uncertain = []
        for id in self.id:
            list_cond = [
                self.get_smiles_graph(cond, id) for cond in self.graph_conditions
            ]
            intersection, first_key_index = self.intersection_of_lists_with_count(
                list_cond, intersection_num
            )
            if intersection and first_key_index is not None:
                match_cond = self.graph_conditions[first_key_index]
                new_graph = [value for value in match_cond if value["R-id"] == id]
                new_graph_uncertain.extend(new_graph)
            else:
                new_graph = [
                    value
                    for value in self.finalgraph_uncertainty
                    if value["R-id"] == id
                ]
                new_graph_uncertain.extend(new_graph)

        return new_graph_uncertain



import pandas as pd

from rdkit import Chem
from joblib import Parallel, delayed
from typing import List
import multiprocessing
import multiprocessing.pool
from synrbl.SynMCSImputer.MissingGraph.find_missing_graphs import FindMissingGraphs
from synrbl.SynMCSImputer.MissingGraph.uncertainty_graph import GraphMissingUncertainty


def find_single_graph_parallel(mcs_mol_list, sorted_reactants_mol_list, n_jobs=4):
    """
    Find missing parts, boundary atoms, and nearest neighbors for a list of
    reactant molecules using a corresponding list of MCS (Maximum Common
    Substructure) molecules in parallel.

    Parameters:
    - mcs_mol_list (list of rdkit.Chem.Mol): List of RDKit molecule objects
        representing the MCS, corresponding to each molecule in
        sorted_reactants_mol_list.
    - sorted_reactants_mol_list (list of rdkit.Chem.Mol): The list of RDKit
        molecule objects to analyze.
    - n_jobs (int): The number of parallel jobs to run. Default is -1, which
        uses all available CPU cores.

    Returns:
    - List of dictionaries, where each dictionary contains:
    - 'smiles' (list of str): SMILES representations of the missing parts for
        each molecule.
    - 'boundary_atoms_products' (list of dict): Lists of boundary atoms for
        each molecule.
    - 'nearest_neighbor_products' (list of dict): Lists of nearest neighbors
        for each molecule.
    - 'issue' (str): Any issues encountered during processing.
    """

    def process_single_pair(reactant_mol, mcs_mol, job_timeout=2):
        output = {
            "smiles": [],
            "boundary_atoms_products": [],
            "nearest_neighbor_products": [],
            "issue": "Find Missing Graph terminated by timeout",
        }
        try:
            pool = multiprocessing.pool.ThreadPool(1)
            async_result = pool.apply_async(
                FindMissingGraphs.find_missing_parts_pairs,
                (
                    reactant_mol,
                    mcs_mol,
                ),
            )
            result = async_result.get(job_timeout)
            pool.terminate()  # Terminate the pool to release resources
            output["smiles"] = [
                Chem.MolToSmiles(mol) if mol is not None else None for mol in result[0]
            ]
            output["boundary_atoms_products"] = result[1]
            output["nearest_neighbor_products"] = result[2]
        except multiprocessing.TimeoutError:
            output["issue"] = "Find Missing Graph terminated by timeout"
        except Exception as e:
            output["issue"] = "Find MCS failed with exception: {}".format(e)
        return output

    results = Parallel(n_jobs=n_jobs, verbose=0)(
        delayed(process_single_pair)(reactant_mol, mcs_mol)
        for reactant_mol, mcs_mol in zip(sorted_reactants_mol_list, mcs_mol_list)
    )
    return results


def find_graph_dict(mcs_dict, n_jobs: int = 4):
    """
    Function to find missing graphs for a given MCS dictionary.
    """
    if len(mcs_dict) == 0:
        return []

    msc_df = pd.DataFrame(mcs_dict)

    mcs_results = msc_df["mcs_results"].to_list()
    sorted_reactants = msc_df["sorted_reactants"].to_list()

    mcs_mol_list = smiles_to_mol_parallel(mcs_results, useSmiles=False)
    sorted_reactants_mol_list = smiles_to_mol_parallel(sorted_reactants, useSmiles=True)

    missing_results = find_single_graph_parallel(
        mcs_mol_list, sorted_reactants_mol_list, n_jobs=n_jobs
    )

    missing_results = GraphMissingUncertainty(missing_results, threshold=2).fit()

    return missing_results


def convert_smiles_to_mols(
    smiles_list: List[str], useSmiles: bool = True
) -> List[Chem.rdchem.Mol]:
    """
    Convert a list of SMILES strings to a list of RDKit molecule objects.

    Args:
        smiles_list: A list containing SMILES strings.

    Returns:
        A list containing RDKit molecule objects.
    """
    mol_list = []
    for smiles in smiles_list:
        try:
            if useSmiles:
                mol = Chem.MolFromSmiles(smiles, sanitize=True)
            else:
                mol = Chem.MolFromSmarts(smiles)
            mol_list.append(mol)
        except Exception:
            mol_list.append(None)
    return mol_list


def smiles_to_mol_parallel(
    smiles_lists: List[List[str]], n_jobs: int = 4, useSmiles: bool = True
) -> List[List[Chem.Mol]]:
    """
    Convert a list of lists of SMILES strings to a list of lists of RDKit
    molecule objects using parallel processing.

    :param smiles_lists: List of lists containing SMILES strings.
    :param n_jobs: The number of jobs to run in parallel. -1 means using
        all processors.
    :return: A list of lists containing RDKit molecule objects.
    """
    mol_lists = Parallel(n_jobs=n_jobs)(
        delayed(convert_smiles_to_mols)(smiles_list, useSmiles)
        for smiles_list in smiles_lists
    )
    return mol_lists



import json
import random

from typing import List, Dict, Set, Any
from typing import Optional, Union, Callable, Tuple
from collections import defaultdict


def save_database(database: list[dict], pathname: str = "./Data/database.json") -> None:
    """
    Save a database (a list of dictionaries) to a JSON file.

    Args:
        database: The database to be saved.
        pathname: The path where the database will be saved.
            Defaults to './Data/database.json'.

    Raises:
        TypeError: If the database is not a list of dictionaries.
        ValueError: If there is an error writing the file.
    """
    if not all(isinstance(item, dict) for item in database):
        raise TypeError("Database should be a list of dictionaries.")

    try:
        with open(pathname, "w") as f:
            json.dump(database, f)
    except IOError as e:
        raise ValueError(f"Error writing to file {pathname}: {e}")


def load_database(pathname: str = "./Data/database.json") -> List[Dict]:
    """
    Load a database (a list of dictionaries) from a JSON file.

    Args:
        pathname: The path from where the database will be loaded.
            Defaults to './Data/database.json'.

    Returns:
        The loaded database.

    Raises:
        ValueError: If there is an error reading the file.
    """
    try:
        with open(pathname, "r") as f:
            database = json.load(f)  # Load the JSON data from the file
        return database
    except IOError as e:
        raise ValueError(f"Error reading to file {pathname}: {e}")


def extract_atomic_elements(rules: List[Dict[str, Dict[str, int]]]) -> Set[str]:
    """
    Extracts the set of all atomic elements from a list of rules.

    Args:
        rules: A list of rules, where each rule is a dictionary representing a
            composition of atomic elements.

    Returns:
        A set of all atomic elements found in the rules.

    Example:
        ```python
        rules = [
            {"Composition": {"A": 1, "B": 1}},
            {"Composition": {"C": 1, "D": 1, "E": 1}},
        ]
        atomic_elements = extract_atomic_elements(rules)
        print(atomic_elements)  # Output: {'A', 'B', 'C', 'D', 'E'}
        ```
    """

    atomic_elements = set()

    for rule in rules:
        atomic_elements.update(rule["Composition"].keys())

    return atomic_elements


def _get_max_comp_len(database: List[Dict]) -> int:
    """
    Determines the maximum length of a composition in the database.

    Args:
        database (List[Dict]): A list of dictionaries representing the database.

    Returns:
        int: The maximum length of a composition in the database.

    Example:
    ```python
    database = [
        {'Composition': ['H2O', 'CO2']},
        {'Composition': ['NaCl', 'H2O']},
        {'Composition': ['C6H12O6', 'H2O']},
    ]

    max_comp_len = _get_max_comp_len(database)
    print(max_comp_len)  # Output: 3
    ```
    """

    max_comp_len = 0  # Initialize the maximum composition length

    # Iterate through each entry in the database
    for entry in database:
        # Check if the current entry's composition length exceeds the maximum
        if max_comp_len < len(entry["Composition"]):
            # Update the maximum composition length if necessary
            max_comp_len = len(entry["Composition"])

    # Return the maximum composition length
    return max_comp_len


def find_shortest_sublists(solution: List[List[Dict]]) -> List[List[Dict]]:
    """
    Find all sublists of dictionaries that have the shortest length.

    Args:
        solution: A list containing lists of dictionaries.

    Returns:
        A list of all sublists with the shortest length.
    """
    if not solution:
        return []

    min_length = min(len(sublist) for sublist in solution)
    shortest_sublists = [sublist for sublist in solution if len(sublist) == min_length]

    return shortest_sublists


def filter_data(
    data: List[Dict[str, Any]],
    unbalance_values: Optional[List[str]] = None,
    formula_key: str = "Diff_formula",
    element_key: Optional[str] = None,
    min_count: int = 0,
    max_count: int = float("inf"),
) -> List[Dict[str, Any]]:
    """
    Filter dictionaries based on a list of unbalance values and element count
    in a specified formula key.

    This function filters the input list of dictionaries based on the specified
    list of unbalance values and the count of a specific element within a given
    formula key. It returns dictionaries that match any of the unbalance
    criteria and where the element count falls within the specified range.

    Args:
        data: A list of dictionaries to be filtered.
        unbalance_values: The values to filter by in the 'Unbalance' key. If
            None, this criterion is ignored.
        formula_key: The key in the dictionaries that contains the element
            counts. Defaults to 'Diff_formula'.
        element_key: The element to filter by in the formula key. If None, this
            criterion is ignored.
        min_count: The minimum allowed count of the element. Defaults to 0.
        max_count: The maximum allowed count of the element.
            Defaults to infinity.

    Returns:
        A list of dictionaries filtered based on the criteria.
    """
    filtered_data = []

    for item in data:
        # Check for unbalance condition
        unbalance_matches = (
            unbalance_values is None or item.get("Unbalance") in unbalance_values
        )

        # Check for element count condition
        if element_key is None:
            element_matches = True
        else:
            element_count = item.get(formula_key, {}).get(element_key, 0)
            element_matches = min_count <= element_count <= max_count

        if unbalance_matches and element_matches:
            filtered_data.append(item)

    return filtered_data


def remove_duplicates_by_key(
    data: List[dict], key_function: Callable[..., Any]
) -> List[dict]:
    """
    Remove duplicate entries from a list based on a unique key for each entry.

    Parameters:
    - `data` (List[dict]): A list of data entries (dictionaries, objects, etc.).
    - `key_function` (Callable[..., Any]): A function that takes an entry from
        `data` and returns a key for duplicate check.

    Returns:
    - `List[dict]`: A list of unique entries, based on the unique keys generated.

    Example:
    >>> data = [
        {"name": "Alice", "age": 30},
        {"name": "Bob", "age": 25},
        {"name": "Alice", "age": 30},
    ]
    >>> remove_duplicates_by_key(data, lambda x: (x["name"], x["age"]))
    [{'name': 'Alice', 'age': 30}, {'name': 'Bob', 'age': 25}]
    """

    seen_keys = set()
    unique_data = []

    for entry in data:
        key = frozenset(key_function(entry))
        if key not in seen_keys:
            seen_keys.add(key)
            unique_data.append(entry)

    return unique_data


def sort_by_key_length(
    data: List[Any], key_function: Callable[[Any], Any]
) -> List[Any]:
    """
    Sort a list of entries based on the length of a specific key.

    Args:
    - data (List[Any]): A list of data entries.
    - key_function (Callable[[Any], Any]): A function that takes an entry from
        `data` and returns a key whose length is to be used for sorting.

    Returns:
    - List[Any]: A list of entries sorted by the length of the specified key.
    """

    return sorted(data, key=lambda x: len(key_function(x)))


def add_missing_key_to_dicts(
    data: List[Dict[str, Dict[str, Any]]],
    dict_key: str,
    missing_key: str,
    default_value: Any,
) -> List[Dict[str, Dict[str, Any]]]:
    """
    Iterates through a list of dictionaries and adds a specified key with a
    default value to a specified dictionary within each main dictionary, if the
    key is not already present. Returns a new list with the updates.

    Args:
        data: A list of dictionaries.
        dict_key: The key in the main dictionaries that points to another
            dictionary where the check should be done.
        missing_key: The key to add if it's not present in the nested dictionary.
        default_value: The default value to assign to the missing key.

    Returns:
        A new list of dictionaries with the missing key added where necessary.

    Example:
        >>> data = [{'Composition': {'A': 1, 'B': 2}}, {'Composition': {'B': 3}}]
        >>> updated_data = add_missing_key_to_dicts(data, 'Composition', 'Q', 0)
        >>> updated_data
        [{'Composition': {'A': 1, 'B': 2, 'Q': 0}}, {'Composition': {'B': 3, 'Q': 0}}]
    """

    updated_data = []

    for entry in data:
        # Create a copy of the entry to avoid modifying the original data
        updated_entry = entry.copy()

        # Check if the missing key is not in the specified dictionary
        if missing_key not in updated_entry.get(dict_key, {}):
            # Ensure the dict_key points to a dictionary
            if dict_key not in updated_entry:
                updated_entry[dict_key] = {}

            # Add the missing key with the default value
            updated_entry[dict_key][missing_key] = default_value

        updated_data.append(updated_entry)

    return updated_data


def extract_results_by_key(
    data: List[Dict[str, any]], key: str = "new_reaction"
) -> Tuple[List[Dict[str, any]], List[Dict[str, any]]]:
    """
    Separate dictionaries from a list into two lists based on the presence of a
    specific key.

    Args:
        data: A list of dictionaries to be separated.
        key: The key to check for in each dictionary.
            Defaults to 'new_reaction'.

    Returns:
        A tuple of two lists:
            - The first list contains dictionaries that have the specified key.
            - The second list contains dictionaries that do not have the
                specified key.
    """

    with_key: List[Dict[str, any]] = []
    without_key: List[Dict[str, any]] = []

    # Separate dictionaries based on the presence of the key
    for item in data:
        if key in item:
            with_key.append(item)
        else:
            without_key.append(item)

    return with_key, without_key


def get_random_samples_by_key(
    data: List[Dict[str, Any]],
    stratify_key: str,
    num_samples_per_group: int = 1,
    random_seed: int = None,
) -> List[Dict[str, Any]]:
    """
    Get random samples from data, grouped by a specified key.

    Parameters:
    - data: List of dictionaries containing various keys.
    - stratify_key: The key used for stratifying the data.
    - num_samples_per_group: Number of random samples to draw from each
        unique group.
    - random_seed: Seed for the random number generator for reproducibility.

    Returns:
    - A list of randomly selected samples from each unique group based on
        the stratify_key.
    """
    if random_seed is not None:
        random.seed(random_seed)

    # Function to create a sortable key from the item
    def sortable_key(item: Dict[str, Any]) -> Union[str, tuple]:
        key_value = item.get(stratify_key, None)
        if isinstance(key_value, dict):
            return tuple(sorted(key_value.items()))
        elif isinstance(key_value, list):
            return tuple(sorted(key_value))
        else:
            return key_value

    # Group data by the specified stratify_key
    grouped_data = defaultdict(list)
    for item in data:
        grouped_key = sortable_key(item)
        grouped_data[grouped_key].append(item)

    # Select random samples from each group
    random_samples = []
    for _, items in grouped_data.items():
        if len(items) >= num_samples_per_group:
            random_samples.extend(random.sample(items, num_samples_per_group))
        else:
            # If there aren't enough items, take all available
            random_samples.extend(items)

    return random_samples



from synrbl.SynMCSImputer.structure import CompoundSet
from synrbl.SynMCSImputer.utils import is_carbon_balanced
from synrbl.SynMCSImputer.merge import merge
from rdkit.rdBase import BlockLogs


def build_compounds(data_dict) -> CompoundSet:
    src_smiles = data_dict["sorted_reactants"]
    smiles = data_dict["smiles"]
    boundaries = data_dict["boundary_atoms_products"]
    neighbors = data_dict["nearest_neighbor_products"]
    mcs_results = data_dict["mcs_results"]
    n = len(smiles)
    if n != len(src_smiles):
        raise ValueError(
            "Smiles and sorted reactants are not of the same length. ({} != {})".format(
                len(smiles), len(src_smiles)
            )
        )
    if n != len(boundaries) or n != len(neighbors):
        raise ValueError(
            "Boundaries and nearest neighbors must be of same length as compounds."
        )
    if n != len(mcs_results):
        raise ValueError("MCS results must be of same length as compounds.")
    cset = CompoundSet()
    for s, ss, b, n, mcs in zip(smiles, src_smiles, boundaries, neighbors, mcs_results):
        if s is None:
            if mcs == "":
                # TODO use compound rule for that
                if ss == "O":
                    # water is not catalyst -> binds to other compound
                    c = cset.add_compound(ss, src_mol=ss)
                    c.add_boundary(0, symbol="O")
                else:
                    # catalysis compound
                    c = cset.add_compound(ss, src_mol=ss)
            else:
                # empty compound
                pass
        else:
            c = cset.add_compound(s, src_mol=ss)
            if len(b) != len(n):
                raise ValueError(
                    (
                        "Boundary and neighbor missmatch. "
                        + "(boundary={}, neighbor={})"
                    ).format(b, n)
                )
            for bi, ni in zip(b, n):
                bi_s, bi_i = list(bi.items())[0]
                ni_s, ni_i = list(ni.items())[0]
                c.add_boundary(
                    bi_i, symbol=bi_s, neighbor_index=ni_i, neighbor_symbol=ni_s
                )
    return cset


def impute_reaction(
    reaction_dict,
    reaction_col,
    issue_col,
    carbon_balance_col,
    mcs_data_col,
    smiles_standardizer=[],
):
    issue = reaction_dict[issue_col] if issue_col in reaction_dict.keys() else ""
    if issue != "":
        raise ValueError("Skip reaction because of previous issue.\n" + issue)
    compound_set = build_compounds(reaction_dict[mcs_data_col])
    if len(compound_set) == 0:
        raise ValueError("Empty compound set.")
    merge_result = merge(compound_set)
    carbon_balance = reaction_dict[carbon_balance_col]
    if carbon_balance == "reactants":
        # Imputing reactant side carbon imbalance is not (yet) supported
        raise ValueError("Skipped because of reactants imbalance.")
    elif carbon_balance in ["products", "balanced"]:
        merged_smiles = merge_result.smiles
        for standardizer in smiles_standardizer:
            merged_smiles = standardizer(merged_smiles)
        imputed_reaction = "{}.{}".format(reaction_dict[reaction_col], merged_smiles)
    else:
        raise ValueError(
            "Invalid value '{}' for carbon balance.".format(carbon_balance)
        )
    rules = [r.name for r in merge_result.rules]
    is_balanced = is_carbon_balanced(imputed_reaction)
    if not is_balanced:
        raise RuntimeError(
            (
                "Failed to impute the correct structure. "
                + "Carbon atom count in reactants and products does not match."
            )
        )
    return imputed_reaction, rules


class MCSBasedMethod:
    def __init__(
        self,
        reaction_col,
        output_col,
        mcs_data_col="mcs",
        issue_col="issue",
        rules_col="rules",
        carbon_balance_col="carbon_balance_check",
        smiles_standardizer=[],
    ):
        self.reaction_col = reaction_col
        self.output_col = output_col
        self.mcs_data_col = mcs_data_col
        self.issue_col = issue_col
        self.rules_col = rules_col
        self.carbon_balance_col = carbon_balance_col
        self.smiles_standardizer = smiles_standardizer

    def run(self, reactions: list[dict], stats=None):
        mcs_applied = 0
        mcs_solved = 0
        block_logs = BlockLogs()
        for reaction in reactions:
            if self.mcs_data_col not in reaction.keys():
                continue
            mcs_applied += 1
            if reaction[self.mcs_data_col] is None:
                continue
            try:
                result, rules = impute_reaction(
                    reaction,
                    mcs_data_col=self.mcs_data_col,
                    reaction_col=self.reaction_col,
                    issue_col=self.issue_col,
                    carbon_balance_col=self.carbon_balance_col,
                    smiles_standardizer=self.smiles_standardizer,
                )
                reaction[self.output_col] = result
                reaction[self.rules_col] = rules
                mcs_solved += 1
            except Exception as e:
                reaction[self.issue_col] = str(e)

        del block_logs
        if stats is not None:
            stats["mcs_applied"] = mcs_applied
            stats["mcs_solved"] = mcs_solved
        return reactions



import unittest

from synrbl.SynMCSImputer.SubStructure.extract_common_mcs import ExtractMCS


class TestExtractMCS(unittest.TestCase):
    def setUp(self):
        self.extractor = ExtractMCS()

    def test_get_num_atoms(self):
        # Test calculation of the number of atoms in a molecule
        smiles = "CCO"
        num_atoms = self.extractor.get_num_atoms(smiles)
        self.assertEqual(num_atoms, 3)

    def test_calculate_total_number_atoms_mcs_parallel(self):
        # Test parallel calculation of total number of atoms in MCS results
        condition = [{"mcs_results": ["CCO", "CC"]}]
        total_atoms = self.extractor.calculate_total_number_atoms_mcs_parallel(
            condition
        )
        self.assertEqual(total_atoms, [5])

    def test_get_popular_elements_from_list(self):
        # Test finding the most popular elements
        elements_list = ["H", "C", "C", "O"]
        popular_elements = self.extractor.get_popular_elements_from_list(elements_list)
        self.assertIn("C", popular_elements)

    def test_get_top_n_common_elements(self):
        # Test finding the top n most common elements
        elements_list = [{"H", "C"}, {"C", "O"}]
        top_elements = self.extractor.get_top_n_common_elements(elements_list, top_n=2)
        self.assertEqual(len(top_elements), 2)

    def test_calculate_corrected_individual_overlap_percentage(self):
        # Test calculating corrected individual overlap percentage
        conditions = [[{"mcs_results": ["CCO"]}], [{"mcs_results": ["CC"]}]]
        (
            overlap_percentages,
            reference_results_list,
        ) = self.extractor.calculate_corrected_individual_overlap_percentage(
            *conditions
        )
        self.assertIsNotNone(overlap_percentages)

    def test_extract_common_mcs_index(self):
        # Test extracting common MCS index
        conditions = [[{"mcs_results": ["CCO"]}], [{"mcs_results": ["CC"]}]]
        (
            threshold_index,
            reference_results_list,
        ) = self.extractor.extract_common_mcs_index(0, 100, *conditions)
        self.assertIsNotNone(threshold_index)


if __name__ == "__main__":
    unittest.main()



import pandas as pd
import rdkit.Chem as Chem

from joblib import Parallel, delayed
from synrbl.rsmi_utils import save_database

from typing import Tuple


class RSMIProcessing:
    """
    A class to process reaction SMILES (RSMI) data.

    Parameters
    ----------
    rsmi : str, optional
        The reaction SMILES string to be processed.
    data : DataFrame, optional
        The DataFrame containing the RSMI data.
    rsmi_col : str, optional
        The column name in the DataFrame that contains the RSMI data.
    symbol : str, optional
        The symbol that separates reactants and products in the RSMI string
        (default is '>>').
    n_jobs : int, optional
        The number of jobs to run in parallel (default is 10).
    verbose : int, optional
        The verbosity level (default is 1).
    parallel : bool, optional
        Whether to use parallel processing (default is True).
    save_json : bool, optional
        Whether to save the processed data to a JSON file (default is True).
    save_path_name : str, optional
        The path and name of the JSON file to save (default is 'reaction.json.gz').
    orient : str, optional
        The orientation of the JSON file (default is 'records').
    compression : str, optional
        The compression type for the JSON file (default is 'gzip').

    Example
    -------
    # Create an instance of the RSMIProcessing class with a DataFrame
    >>> data = pd.DataFrame({'rsmi': ['CCO>>CCOCC', 'CC>>C']})
    >>> processor = RSMIProcessing(data=data, rsmi_col='rsmi', parallel=False)
    # Split the RSMI data into reactants and products
    >>> processed_data = processor.data_splitter()
    # Display the processed data
    >>> print(processed_data)
    """

    def __init__(
        self,
        reaction_smiles: str = None,
        data: pd.DataFrame = None,
        data_name: str = "USPTO_50K",
        index_col: str = "R-id",
        rsmi_col: str = None,
        symbol: str = ">>",
        n_jobs: int = 4,
        drop_duplicates: bool = True,
        verbose: int = 1,
        parallel: bool = True,
        save_json: bool = False,
        save_path_name: str = "reaction.json.gz",
        orient: str = "records",
        compression: str = "gzip",
    ) -> None:
        self.reaction_smiles = reaction_smiles
        self.symbol = symbol
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.parallel = parallel
        self.data = data
        self.data_name = data_name
        self.index_col = index_col
        self.drop_duplicates = drop_duplicates
        self.rsmi_col = rsmi_col
        self.save_json = save_json
        self.save_path_name = save_path_name
        self.orient = orient
        self.compression = compression

    @staticmethod
    def smi_splitter(rsmi: str, symbol: str = ">>") -> Tuple[str, str]:
        """
        Split a RSMI string into reactants and products.

        Returns
        -------
        tuple
            The reactants and products as separate strings, if the RSMI string
            can be parsed.
        str
            A message indicating the RSMI string can't be parsed, if applicable.
        """

        # Check if the RSMI string can be parsed
        if RSMIProcessing.can_parse(rsmi, symbol):
            # Split the RSMI string into reactants and products
            return rsmi.split(symbol)
        else:
            # Return a message if the RSMI string cannot be parsed
            return "Can't parse"

    def data_splitter(self) -> pd.DataFrame:
        """
        Split the RSMI data in the DataFrame into reactants and products.

        Returns
        -------
        DataFrame
            The processed DataFrame with separate columns for reactants and products.

        Example
        -------
        >>> data = pd.DataFrame({'rsmi': ['C>>CC', 'CC>>CCC']})
        >>> processor = RSMIProcessing(data=data, rsmi_col='rsmi', parallel=False)
        >>> processed_data = processor.data_splitter()
        >>> print(processed_data)
        """
        # Check if parallel processing is enabled
        if self.parallel:
            # Use joblib's Parallel to concurrently process each RSMI string in
            # the DataFrame 'can_parse' function is applied to each RSMI string
            # to check if it's parsable
            parsable = Parallel(n_jobs=self.n_jobs, verbose=self.verbose)(
                delayed(RSMIProcessing.can_parse)(rsmi)
                for rsmi in self.data[self.rsmi_col].values
            )
            # Filter the data to include only parsable RSMI strings
            self.data = self.data[parsable]
        else:
            # If parallel processing is not enabled, apply 'can_parse'
            # function sequentially to each RSMI string in the DataFrame and
            # filter parsable ones
            self.data = self.data[
                self.data[self.rsmi_col].apply(RSMIProcessing.can_parse)
            ]

        # Split each RSMI string in the DataFrame into reactants and products
        # 'expand=True' splits the string into separate columns
        split_smiles = self.data[self.rsmi_col].str.split(self.symbol, expand=True)

        # Assign the first part of the split (reactants) to a new column in
        # the DataFrame
        self.data["reactants"] = split_smiles[0]
        # Assign the second part of the split (products) to another new column
        self.data["products"] = split_smiles[1]

        if self.drop_duplicates:
            self.data.drop_duplicates(subset=self.rsmi_col, inplace=True)
        self.data.reset_index(drop=True, inplace=True)
        self.data[self.index_col] = [
            self.data_name + "_" if self.data_name is not None else "" + str(i)
            for i in self.data.index
        ]
        # Check if there's a need to save the processed data to a JSON file
        if self.save_json:
            data = self.data.to_dict(orient=self.orient)
            save_database(data, self.save_path_name)
            # Save the DataFrame to a JSON file with specified format
            # and compression

        # Return the processed DataFrame with separate columns for reactants
        # and products
        return self.data

    @staticmethod
    def can_parse(rsmi: str, symbol: str = ">>") -> bool:
        """
        Check if a RSMI string can be parsed into reactants and products.

        Parameters
        ----------
        rsmi : str
            The reaction smiles (RSMI) string to be checked for parsability.
        symbol : str, optional
            The symbol used to separate reactants and products in the RSMI
            string (default is '>>').

        Returns
        -------
        bool
            True if the RSMI string can be parsed into valid reactant and
            product SMILES, False otherwise.
        """

        # Split the RSMI string into reactants and products using the
        # provided symbol
        react, prod = rsmi.split(symbol)

        # Check if both reactants and products can be converted into RDKit
        # molecule objects
        return (
            Chem.MolFromSmiles(prod) is not None
            and Chem.MolFromSmiles(react) is not None
        )



import unittest
from synrbl.SynChemImputer.post_process import PostProcess


class TestPostProcess(unittest.TestCase):
    def setUp(self):
        self.data = [
            {
                "R-id": 1,
                "reactions": "CC(=O)C.[H].[H]>>CC(O)C",
            },
            {
                "R-id": 2,
                "reactions": "CCO.[O]>>CC=O",
            },
            {
                "R-id": 3,
                "reactions": "C=CC>>C1CC1",
            },
        ]

    def test_label_reactions(self):
        post_process = PostProcess()
        result = post_process.label_reactions(self.data[0])
        self.assertEqual(result["label"], "Reduction")
        result = post_process.label_reactions(self.data[1])
        self.assertEqual(result["label"], "Oxidation")
        result = post_process.label_reactions(self.data[2])
        self.assertEqual(result["label"], "unspecified")

    def test_fit(self):
        post_process = PostProcess(n_jobs=4, verbose=1)
        result = post_process.fit(self.data)
        self.assertEqual(result[1]["curated_reaction"], "CC(=O)C.[HH]>>CC(O)C")
        self.assertEqual(
            result[2]["curated_reaction"],
            (
                "CCO.O=[Cr](Cl)(-[O-])=O.c1cc[nH+]cc1>>"
                + "CC=O.O=[Cr](O)O.c1cc[nH+]cc1.[Cl-]"
            ),
        )


if __name__ == "__main__":
    unittest.main()



import logging

from synrbl.SynMCSImputer.SubStructure.mcs_process import ensemble_mcs
from synrbl.SynMCSImputer.SubStructure.extract_common_mcs import ExtractMCS
from synrbl.SynMCSImputer.MissingGraph.find_graph_dict import find_graph_dict

logger = logging.getLogger(__name__)


class MCSSearch:
    def __init__(
        self,
        id_col,
        solved_col="solved",
        mcs_data_col="mcs",
        issue_col="issue",
        n_jobs=-1,
    ):
        self.id_col = id_col
        self.solved_col = solved_col
        self.mcs_data_col = mcs_data_col
        self.issue_col = issue_col
        self.n_jobs = n_jobs

        self.conditions = [
            {
                "RingMatchesRingOnly": True,
                "CompleteRingsOnly": True,
                "method": "MCIS",
                "sort": "MCIS",
                "ignore_bond_order": True,
            },
            # {
            #     "RingMatchesRingOnly": True,
            #     "CompleteRingsOnly": True,
            #     "method": "MCIS",
            #     "sort": "MCIS",
            #     "ignore_bond_order": False,
            # },
            {
                "RingMatchesRingOnly": False,
                "CompleteRingsOnly": False,
                "method": "MCIS",
                "sort": "MCIS",
                "ignore_bond_order": True,
            },
            # {
            #     "RingMatchesRingOnly": False,
            #     "CompleteRingsOnly": False,
            #     "method": "MCIS",
            #     "sort": "MCIS",
            #     "ignore_bond_order": False,
            # },
            {"method": "MCES", "sort": "MCES"},
        ]

    def find(self, reactions):
        id2idx_map = {}
        mcs_reactions = []
        for idx, reaction in enumerate(reactions):
            if reaction[self.solved_col]:
                continue
            id2idx_map[reaction[self.id_col]] = idx
            reaction[self.mcs_data_col] = None
            reaction[self.issue_col] = "No MCS identified."
            mcs_reactions.append(reaction)

        if len(mcs_reactions) == 0:
            return reactions

        logger.info(
            "Find maximum-common-substructure for {} reactions.".format(
                len(mcs_reactions)
            )
        )

        condition_results = ensemble_mcs(
            mcs_reactions,
            self.conditions,
            id_col=self.id_col,
            issue_col=self.issue_col,
            n_jobs=self.n_jobs,
        )

        largest_conditions = ExtractMCS.get_largest_condition(*condition_results)

        mcs_results = find_graph_dict(largest_conditions, n_jobs=self.n_jobs)

        assert len(largest_conditions) == len(mcs_results)
        for largest_condition, mcs_result in zip(largest_conditions, mcs_results):
            _id = largest_condition[self.id_col]
            _idx = id2idx_map[_id]
            for k, v in largest_condition.items():
                mcs_result[k] = v
            reactions[_idx][self.mcs_data_col] = mcs_result
            reactions[_idx][self.issue_col] = mcs_result[self.issue_col]

        return reactions



import copy
import re
from rdkit import Chem
from typing import List, Dict, Pattern, Any, Optional, Tuple


class RuleConstraint:
    """
    A class for applying specific chemical reaction constraints and modifications.

    Attributes
    ----------
    list_dict : list of dict
        A list containing chemical reaction data.
    ban_atoms : list of str
        A list of SMILES strings representing banned atoms or molecules in reactions.

    Methods
    -------
    fit():
        Applies oxidation rules to modify reactions and filters out reactions
        with banned atoms.
    """

    def __init__(
        self,
        list_dict: List[Dict[str, Any]],
        ban_atoms: Optional[List[str]] = None,
        ban_atoms_reactants: Optional[List[str]] = None,
    ) -> None:
        """
        Initializes the RuleConstraint class with a list of chemical reactions
        and optional banned atoms.

        Parameters
        ----------
        list_dict : List[Dict[str, Any]]
            A list of dictionaries, each representing a chemical reaction.
        ban_atoms : Optional[List[str]], optional
            A list of SMILES strings for atoms or molecules to be banned from
            the reactions. Defaults to a predefined list.
        """
        self.list_dict = copy.deepcopy(list_dict)
        self.ban_atoms = ban_atoms or [
            "[H]",
            "[O].[O]",
            "F-F",
            "Cl-Cl",
            "Br-Br",
            "I-I",
            "Cl-Br",
            "Cl-I",
            "Br-I",
        ]
        self.ban_atoms = [Chem.CanonSmiles(atom) for atom in self.ban_atoms]
        self.ban_pattern = re.compile("|".join(map(re.escape, self.ban_atoms)))
        self.ban_atoms_reactants = ban_atoms_reactants or [".[H]"]
        self.ban_pattern_reactants = re.compile(
            "|".join(map(re.escape, self.ban_atoms_reactants))
        )

    @staticmethod
    def reduction_oxidation_rules_modify(
        data: List[Dict[str, str]]
    ) -> List[Dict[str, str]]:
        """
        Modify the oxidation rules in the given data.

        Args:
            data (List[Dict[str, str]]): The input data containing oxidation rules.

        Returns:
            List[Dict[str, str]]: The modified data with updated oxidation rules.
        """
        modified_data = []
        for entry in data:
            if ".[H]" in entry["products"]:
                reactants = entry["reactants"].split(".")
                reactants = [
                    RuleConstraint.remove_atom_mapping(smiles) for smiles in reactants
                ]
                no_constraint = ["[Na]", "[K]", "[Li]", "[H-]"]
                contains_no_constraint = RuleConstraint.check_no_constraint(
                    reactants, no_constraint
                )
                if contains_no_constraint:
                    pass
                else:
                    if RuleConstraint.check_even(entry, "products", "[H]", "."):
                        hydrogen_count = entry["products"].count(".[H]")
                        hydrogen_count = int(hydrogen_count / 2)
                        entry["products"] = entry["products"].replace(".[H]", "")
                        entry["reactants"] += ".[O]" * hydrogen_count

                        if entry["products"]:
                            entry["products"] += ".O" * hydrogen_count
                        else:
                            entry["products"] = "O" * hydrogen_count
                    else:
                        pass

            if ".[O]" in entry["products"]:
                if RuleConstraint.check_even(entry, "products", "[O]", "."):
                    pass
                else:
                    oxygen_count = int(entry["products"].count(".[O]"))
                    entry["products"] = entry["products"].replace(".[O]", "")
                    entry["reactants"] += ".[H].[H]" * oxygen_count

                    if entry["products"]:
                        entry["products"] += ".O" * oxygen_count
                    else:
                        entry["products"] = "O" * oxygen_count

            elif ".OO" in entry["products"]:
                entry["products"] = entry["products"].replace(".OO", "")
                entry["reactants"] += ".[H].[H]"

                if entry["products"]:
                    entry["products"] += ".O.O"
                else:
                    entry["products"] = "O.O"

            new_reaction = f"{entry['reactants']}>>{entry['products']}"
            entry["new_reaction"] = new_reaction
            modified_data.append(entry)

        return modified_data

    @staticmethod
    def remove_banned_reactions(
        reaction_list: List[Dict[str, str]],
        ban_pattern: Pattern,
        ban_pattern_reactants: Pattern,
    ) -> Tuple[List[Dict[str, str]], List[Dict[str, str]]]:
        """
        Filters out reactions that contain banned atoms in their products.

        Parameters:
        reaction_list (List[Dict[str, str]]): A list containing reaction data.
        ban_pattern (Pattern): A compiled regular expression pattern that
            matches any of the banned atoms.

        Returns:
        Tuple[List[Dict[str, str]], List[Dict[str, str]]]: The filtered list
            of reactions without and with banned atoms in their products.
        """
        filtered_reactions = [
            reaction
            for reaction in reaction_list
            if not ban_pattern.search(reaction.get("products", ""))
        ]
        reactions_with_banned_atoms = [
            reaction
            for reaction in reaction_list
            if ban_pattern.search(reaction.get("products", ""))
        ]

        # check validity of reactants, should not contain single [H] or single [O]
        filtered_reactions = [
            reaction
            for reaction in filtered_reactions
            if len(re.findall(ban_pattern_reactants, reaction.get("reactants", ""))) % 2
            == 0
        ]
        reactions_with_banned_atoms_reactants = [
            reaction
            for reaction in reaction_list
            if len(re.findall(ban_pattern_reactants, reaction.get("reactants", ""))) % 2
            != 0
        ]
        reactions_with_banned_atoms.extend(reactions_with_banned_atoms_reactants)
        return filtered_reactions, reactions_with_banned_atoms

    @staticmethod
    def check_even(data_dict: dict, key: str, frag: str, symbol: str = ">>") -> bool:
        """
        Check if the number of occurrences of a fragment in a list of SMILES
        strings is even.

        Args:
            data_dict (dict): A dictionary containing SMILES strings.
            key (str): The key to access the relevant value in the dictionary.
            frag (str): The fragment to count in the SMILES strings.
            symbol (str, optional): The delimiter used in the SMILES strings.
                Defaults to '>>'.

        Returns:
            bool: True if the number of occurrences of the fragment is even,
                False otherwise.
        """
        smiles_list = data_dict[key].split(symbol)
        count = smiles_list.count(frag)
        return count % 2 == 0

    def fit(self) -> List[Dict[str, Any]]:
        """
        Applies oxidation modification rules and filters out reactions with
        banned atoms.

        Returns
        -------
        List[Dict[str, Any]]
            The modified and filtered list of chemical reactions.
        """
        data_modified = self.reduction_oxidation_rules_modify(self.list_dict)
        return self.remove_banned_reactions(
            data_modified, self.ban_pattern, self.ban_pattern_reactants
        )

    @staticmethod
    def remove_atom_mapping(smiles: str) -> str:
        """
        Remove atom mapping numbers from a SMILES string.

        Atom mappings are typically represented by numbers following a
        colon (':') after the atom symbol. This function removes these mappings
        to return a SMILES string without them.

        Args:
            smiles (str): A SMILES string with atom mappings.

        Returns:
            str: A SMILES string without atom mappings.
        """
        # Regular expression to find and remove atom mappings
        # (numbers following a colon)
        mapping_pattern = re.compile(r":\d+")
        return mapping_pattern.sub("", smiles)

    @staticmethod
    def check_no_constraint(reactants, no_constraint):
        """
        Check if any elements in the no_constraint list match exactly with any
        elements in the reactants list.

        Args:
            reactants (List[str]): A list of reactant elements or compounds.
            no_constraint (List[str]): A list of elements to check against in
                the reactants list.

        Returns:
            bool: True if any element from no_constraint is found in reactants,
                False otherwise.
        """
        # Convert lists to sets for efficient membership testing
        reactants_set = set(reactants)
        no_constraint_set = set(no_constraint)

        # Check for intersection
        return not reactants_set.isdisjoint(no_constraint_set)



import unittest
from pathlib import Path

from unittest.mock import patch
from synrbl.SynMCSImputer.SubStructure.mcs_process import single_mcs_safe, ensemble_mcs


class TestMCSFunctions(unittest.TestCase):
    def setUp(self):
        # Example reaction data for testing
        self.sample_reaction_data = {
            "R-id": "example_id",
            "reactants": "CCO.C[O]",
            "products": "CCO",
            "carbon_balance_check": "balanced",
        }
        self.root_dir = Path(".")
        self.conditions = [
            {
                "RingMatchesRingOnly": True,
                "CompleteRingsOnly": True,
                "method": "MCIS",
                "sort": "MCIS",
                "ignore_bond_order": True,
            }
        ]

    @patch(
        "synrbl.SynMCSImputer.SubStructure.mcs_graph_detector.MCSMissingGraphAnalyzer.fit"
    )
    def test_single_mcs(self, mock_fit):
        # Mocking MCSMissingGraphAnalyzer.fit to return predefined values
        mock_fit.return_value = ([], [], [], None)

        result = single_mcs_safe(self.sample_reaction_data, id_col="R-id")
        self.assertEqual(result["R-id"], "example_id")
        self.assertIsInstance(result["mcs_results"], list)
        self.assertIsInstance(result["sorted_reactants"], list)


if __name__ == "__main__":
    unittest.main()


def test_timeout():
    conditions = [
        {
            "RingMatchesRingOnly": True,
            "CompleteRingsOnly": True,
            "method": "MCIS",
            "sort": "MCIS",
            "ignore_bond_order": True,
            "maxNodes": 1000,
        }
    ]
    data = [
        {
            "id": 0,
            "carbon_balance_check": "products",
            "reactants": "[Br]" + 100 * "[Si](C)(C)O" + "[Si][Br]",
            "products": "O" + 100 * "[Si](C)(C)O" + "[Si]O",
        }
    ]
    result = ensemble_mcs(data, conditions, n_jobs=2)
    assert "timeout" in result[0][0]["issue"]
    assert [] == result[0][0]["mcs_results"]



import re
import numpy as np

import rdkit.Chem as Chem
import rdkit.DataStructs as DataStructs
import rdkit.Chem.rdFingerprintGenerator as rdFingerprintGenerator
import rdkit.Chem.AllChem as AllChem
import rdkit.Chem.rdmolfiles as rdmolfiles
from collections import Counter
from typing import List, Dict, Tuple, Optional, Union
from fgutils import FGQuery


class CheckCarbonBalance:
    def __init__(
        self, reactions_data: List[Dict[str, str]], rsmi_col="reactions", symbol=">>"
    ):
        """
        Initialize the CheckCarbonBalance class with reaction data.

        Parameters:
        reactions_data (List[Dict[str, str]]): A list of dictionaries, each
            containing reaction information.
        """
        self.reactions_data = reactions_data
        self.rsmi_col = rsmi_col
        self.symbol = symbol

    @staticmethod
    def count_carbon_atoms(smiles: str) -> int:
        """
        Count the number of carbon atoms in a molecule represented by a SMILES
        string.

        Parameters:
        smiles (str): A SMILES string.

        Returns:
        int: The number of carbon atoms in the molecule. Returns 0 if the
            SMILES string is invalid.
        """
        mol = Chem.MolFromSmiles(smiles)
        return (
            sum(1 for atom in mol.GetAtoms() if atom.GetSymbol() == "C") if mol else 0
        )

    def check_carbon_balance(self) -> None:
        """
        Check and update the carbon balance status for each reaction in the
        reactions data.

        The method updates each reaction dictionary in the reactions data with
        a new key 'carbon_balance_check'. This key will have the value
        'products' if the number of carbon atoms in the products is greater
        than or equal to the reactants, and 'reactants' otherwise.
        """
        for reaction in self.reactions_data:
            try:
                reactants_smiles, products_smiles = reaction[self.rsmi_col].split(
                    self.symbol
                )
                reactants_carbon = sum(
                    self.count_carbon_atoms(smiles)
                    for smiles in reactants_smiles.split(".")
                )
                products_carbon = sum(
                    self.count_carbon_atoms(smiles)
                    for smiles in products_smiles.split(".")
                )

                if reactants_carbon >= products_carbon:
                    reaction["carbon_balance_check"] = "products"
                else:
                    reaction["carbon_balance_check"] = "reactants"
            except KeyError as e:
                print(f"Key error: {e}")
            except ValueError as e:
                print(f"Value error: {e}")

    def is_carbon_balance(self) -> None:
        """
        Check and update the carbon balance status for each reaction in the
        reactions data.

        The method updates each reaction dictionary in the reactions data with
        a new key 'carbon_balance_check'. This key will have the value
        'products' if the number of carbon atoms in the products is greater
        than or equal to the reactants, and 'reactants' otherwise.
        """
        for reaction in self.reactions_data:
            try:
                reactants_smiles, products_smiles = reaction[self.rsmi_col].split(
                    self.symbol
                )
                reactants_carbon = sum(
                    self.count_carbon_atoms(smiles)
                    for smiles in reactants_smiles.split(".")
                )
                products_carbon = sum(
                    self.count_carbon_atoms(smiles)
                    for smiles in products_smiles.split(".")
                )
                reaction["is_carbon_balance"] = reactants_carbon == products_carbon

            except KeyError as e:
                print(f"Key error: {e}")
            except ValueError as e:
                print(f"Value error: {e}")


def calculate_net_charge(sublist: list[dict[str, Union[str, int]]]) -> int:
    """
    Calculate the net charge from a list of molecules represented as SMILES
    strings.

    Args:
        sublist: A list of dictionaries, each with a 'smiles' string and a
            'Ratio' integer.

    Returns:
        The net charge of the sublist as an integer.
    """
    total_charge = 0
    for item in sublist:
        if "smiles" in item and "Ratio" in item:
            mol = Chem.MolFromSmiles(item["smiles"])
            if mol:
                charge = (
                    sum(abs(atom.GetFormalCharge()) for atom in mol.GetAtoms())
                    * item["Ratio"]
                )
                total_charge += charge
    return total_charge


def remove_atom_mapping(smiles: str) -> str:
    pattern = re.compile(r":\d+")
    smiles = pattern.sub("", smiles)
    pattern = re.compile(r"\[(?P<atom>(B|C|N|O|P|S|F|Cl|Br|I){1,2})(?:H\d?)?\]")
    smiles = pattern.sub(r"\g<atom>", smiles)
    return smiles


def remove_stereo_chemistry(smiles: str) -> str:
    pattern = re.compile(r"\[(?P<atom>(\w+))@+\w+\]")
    smiles = pattern.sub(r"[\g<atom>]", smiles)
    return smiles


def count_atoms(smiles: str) -> int:
    pattern = re.compile(r"(B|C|N|O|P|S|F|Cl|Br|I|c|n|o)")
    return len(pattern.findall(smiles))


def normalize_smiles(smiles: str) -> str:
    smiles = remove_stereo_chemistry(smiles)
    if ">>" in smiles:
        return ">>".join([normalize_smiles(t) for t in smiles.split(">>")])
    elif "." in smiles:
        token = smiles.split(".")
        token = [normalize_smiles(t) for t in token]
        token.sort(
            key=lambda x: (count_atoms(x), sum(ord(c) for c in x)),
            reverse=True,
        )
        return ".".join(token)
    else:
        return Chem.CanonSmiles(remove_atom_mapping(smiles))


def _get_diff_mol(smiles1, smiles2):
    smiles1 = normalize_smiles(smiles1)
    smiles2 = normalize_smiles(smiles2)
    diff_1 = []
    diff_2 = []
    for s1, s2 in zip(smiles1.split("."), smiles2.split(".")):
        if s1 != s2:
            diff_1.append(s1)
            diff_2.append(s2)
    mol_1, mol_2 = Chem.RWMol(), Chem.RWMol()
    if len(diff_1) > 0:
        s_1 = ".".join(diff_1)
        mol_1 = rdmolfiles.MolFromSmiles(s_1)
    if len(diff_2) > 0:
        s_2 = ".".join(diff_2)
        mol_2 = rdmolfiles.MolFromSmiles(s_2)
    return mol_1, mol_2


def wc_similarity(expected_smiles, result_smiles, method="pathway"):
    """
    Compute the worst case similarity between two SMILES. The method compares
    the molecules that are only present on one sie of the reaction to not bias
    the similarity by the size of the compounds. The similarity is the lower
    value from educt and product similarity.

    Args:
        expected_smiles (str): The expected reaction SMILES.
        result_smiles (str): The actual reaction SMILES.
        method (str): The method used to compute similarity.
            Possible values are: ['pathway', 'ecfp', 'ecfp_inv']
    Returns:
        float: The similarity value between 0 and 1
    """

    def _fp(mol1, mol2):
        sim = 0
        if method == "pathway":
            fpgen = AllChem.GetRDKitFPGenerator(maxPath=5, minPath=1)
            fp1 = fpgen.GetFingerprint(mol1)
            fp2 = fpgen.GetFingerprint(mol2)
            sim = DataStructs.TanimotoSimilarity(fp1, fp2)
        elif method == "ecfp":
            fpgen = rdFingerprintGenerator.GetMorganGenerator(radius=2)
            fp1 = fpgen.GetSparseCountFingerprint(mol1)
            fp2 = fpgen.GetSparseCountFingerprint(mol2)
            sim = DataStructs.DiceSimilarity(fp1, fp2)
        elif method == "ecfp_inv":
            invgen = rdFingerprintGenerator.GetMorganFeatureAtomInvGen()
            ffpgen = rdFingerprintGenerator.GetMorganGenerator(
                radius=2, atomInvariantsGenerator=invgen
            )
            fp1 = ffpgen.GetSparseCountFingerprint(mol1)
            fp2 = ffpgen.GetSparseCountFingerprint(mol2)
            sim = DataStructs.DiceSimilarity(fp1, fp2)
        else:
            raise ValueError("'{}' is not a valid similarity method.".format(method))
        return sim

    exp = normalize_smiles(expected_smiles)
    res = normalize_smiles(result_smiles)
    if exp == res:
        return 1
    exp_e, exp_p = exp.split(">>")
    res_e, res_p = res.split(">>")
    exp_ed, res_ed = _get_diff_mol(exp_e, res_e)
    exp_pd, res_pd = _get_diff_mol(exp_p, res_p)
    return np.min([_fp(exp_ed, res_ed), _fp(exp_pd, res_pd)])


def check_for_isolated_atom(smiles: str, atom: Optional[str] = "H") -> bool:
    """
    Checks if a specified type of isolated atom (hydrogen by default, or oxygen)
    exists in a SMILES string.
    """
    # Pattern to find isolated atoms; not connected to any other atoms
    pattern = rf"\[{atom}\](?![^[]*\])"
    return bool(re.search(pattern, smiles))


def count_radical_atoms(smiles: str, atomic_num: int) -> int:
    """
    Counts isolated radical atoms in SMILES string.
    """
    mol = Chem.MolFromSmiles(smiles)
    radical_count = 0

    # Iterate over all atoms in the molecule
    for atom in mol.GetAtoms():
        if atom.GetAtomicNum() == atomic_num and atom.GetNumRadicalElectrons() > 0:
            # Further check if the atom is isolated (has no neighbors)
            if len(atom.GetNeighbors()) == 0:
                radical_count += 1

    return radical_count


def list_difference(
    list1: List[str], list2: List[str]
) -> Tuple[Dict[str, int], Dict[str, int]]:
    """
    Compares two lists and returns dictionaries that count unique occurrences
    Parameters:
    list1 (List[str]): First list of items for comparison.
    list2 (List[str]): Second list of items for comparison.

    Returns:
    Tuple[Dict[str, int], Dict[str, int]]: A tuple of two dictionaries:
        - First dictionary: Items unique to the first list with their counts.
        - Second dictionary: Items unique to the second list with their counts.
    """
    count1 = Counter(list1)
    count2 = Counter(list2)
    unique_to_list1 = {}
    unique_to_list2 = {}

    for key, count in count1.items():
        if key not in count2:
            unique_to_list1[key] = count
        elif count > count2[key]:
            unique_to_list1[key] = count - count2[key]

    for key, count in count2.items():
        if key not in count1:
            unique_to_list2[key] = count
        elif count > count1[key]:
            unique_to_list2[key] = count - count1[key]

    return unique_to_list1, unique_to_list2


def find_functional_reactivity(reaction_smiles: str) -> Tuple[List[str], List[str]]:
    """
    Analyzes functional groups

    Parameters:
    reaction_smiles (str): SMILES string of the reaction
    Returns:
    Tuple[List[str], List[str]]: Two lists containing unique functional groups
    in reactants and products, respectively.
    """
    query = FGQuery(use_smiles=True)
    reactant, product = reaction_smiles.split(">>")
    fg_reactant = query.get(reactant)
    fg_product = query.get(product)
    fg_reactant = [value[0] for value in fg_reactant]
    fg_product = [value[0] for value in fg_product]
    reactant_fg, product_fg = list_difference(fg_reactant, fg_product)
    return list(reactant_fg.keys()), list(product_fg.keys())



import pandas as pd
import xgboost as xgb
import matplotlib.pyplot as plt
import numpy as np

from copy import deepcopy
from sklearn.preprocessing import LabelEncoder
from xgboost import XGBClassifier
from typing import List, Optional


class FeatureAnalysis:
    def __init__(
        self,
        data: pd.DataFrame,
        target_col: str,
        cols_for_contour: List[List[str]],
        figsize: tuple = (16, 12),
    ):
        """
        Initialize the FeatureAnalysis class.

        Parameters:
        - data (pd.DataFrame): The DataFrame containing the data.
        - target_col (str): The name of the target column.
        - cols_for_contour (List[List[str]]): List of lists containing feature
            pairs for contour plots.
        - figsize (tuple): Figure size for the visualization. Default is (16, 12).
        """
        self.data = data
        self.target_col = target_col
        self.cols_for_contour = cols_for_contour
        self.figsize = figsize

    def feature_importance(self, ax: plt.Axes) -> None:
        """
        Create a feature importance plot using XGBoost.

        Parameters:
        - ax (plt.Axes): The axes on which to plot the feature importance.

        Returns:
        None
        """
        X = self.data.drop(self.target_col, axis=1)
        y = self.data[self.target_col]
        ax.tick_params(axis="x", colors="black")
        ax.tick_params(axis="y", colors="black")
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.spines["left"].set_color("grey")
        ax.spines["bottom"].set_color("grey")
        ax.xaxis.grid(True, which="major", alpha=0.3)

        clf = xgb.XGBClassifier(random_state=42)
        clf.fit(X, y)

        feature_importances = clf.feature_importances_
        importance_df = pd.DataFrame(
            {"Feature": X.columns, "Importance": feature_importances}
        )
        importance_df = importance_df.sort_values(by="Importance", ascending=False)

        ax.set_xlabel("Importance", color="black")
        ax.invert_yaxis()

    def contour_plot(self, features: List[str], ax: plt.Axes) -> None:
        """
        Create a contour plot for a pair of features.

        Parameters:
        - features (List[str]): List of two feature names.
        - ax (plt.Axes): The axes on which to plot the contour plot.

        Returns:
        None
        """
        df = deepcopy(self.data)
        le = LabelEncoder()
        df["Outcome"] = le.fit_transform(df[self.target_col])

        X = df[list(features)]
        y = df["Outcome"]

        model = XGBClassifier(random_state=42)
        model.fit(X, y)

        x_min, x_max = X[features[0]].min() - 1, X[features[0]].max() + 1
        y_min, y_max = X[features[1]].min() - 1, X[features[1]].max() + 1
        xx, yy = np.meshgrid(
            np.linspace(x_min, x_max, 1000), np.linspace(y_min, y_max, 1000)
        )  # Increase the number of points

        Z = model.predict_proba(np.c_[xx.ravel(), yy.ravel()])[:, 1]
        Z = Z.reshape(xx.shape)

        ax.set_xlabel(features[0])
        ax.set_ylabel(features[1])

    def visualize(self, save_path: Optional[str] = None) -> None:
        """
        Visualize feature importance and contour plots.

        Parameters:
        - save_path (Optional[str]): Path to save the figure. If None, the
            figure is not saved. Default is None.

        Returns:
        None
        """
        fig, ax = plt.subplots(1, 1, figsize=self.figsize)

        # Feature importance plot (A)
        self.feature_importance(ax)

        plt.tight_layout()

        if save_path:
            plt.savefig(save_path, transparent=True, bbox_inches="tight")

        plt.show()



from rdkit import Chem
from rdkit.Chem import rdmolops
from rdkit.Chem import AllChem


class MoleculeCurator:
    """
    A class for curating molecules, which includes adding hydrogens to radicals,
    standardizing diazo charges, and manually kekulizing molecules.

    Methods:
    - add_hydrogens_to_radicals: Adds hydrogens to radical sites in a molecule.
    - standardize_diazo_charge: Converts a diazo compound with charged atoms to
        its neutral form.
    - manual_kekulize: Manually kekulizes a molecule to ensure valid structure.
    """

    @staticmethod
    def add_hydrogens_to_radicals(mol: Chem.Mol) -> Chem.Mol:
        """
        Add hydrogen atoms to radical sites in a molecule.

        Args:
        - mol (Chem.Mol): RDKit molecule object.

        Returns:
        - Chem.Mol: The modified molecule with added hydrogens.
        """
        mol_with_h = Chem.RWMol(mol)
        mol_with_h = rdmolops.AddHs(mol_with_h)

        for atom in mol_with_h.GetAtoms():
            num_radical_electrons = atom.GetNumRadicalElectrons()
            if num_radical_electrons > 0:
                atom.SetNumExplicitHs(atom.GetNumExplicitHs() + num_radical_electrons)
                atom.SetNumRadicalElectrons(0)

        curate_mol = Chem.RemoveHs(mol_with_h)
        return curate_mol

    @staticmethod
    def standardize_diazo_charge(mol: Chem.Mol) -> Chem.Mol:
        """
        Convert a diazo compound with charged atoms to its neutral form.

        Args:
        - mol (Chem.Mol): A RDKit molecule object representing a diazo compound.

        Returns:
        - Chem.Mol: The neutralized molecule, or the original molecule if the
            reaction doesn't occur.
        """
        neutral_mol = AllChem.ReactionFromSmarts(
            "[N-]=[NH2+]>>[N:1]#[N:2]"
        ).RunReactants((mol,))
        return neutral_mol[0][0] if neutral_mol else mol

    @staticmethod
    def manual_kekulize(smiles: str) -> Chem.Mol:
        """
        Manually kekulizes a molecule represented by a SMILES string.

        Args:
        - smiles (str): A SMILES string representing the molecule.

        Returns:
        - Chem.Mol: The kekulized molecule, or None if it's not possible.
        """

        # Function to add a hydrogen to a specific atom in a molecule
        def add_hydrogen(mol, atom_index):
            edited_mol = Chem.RWMol(mol)
            atom = edited_mol.GetAtomWithIdx(atom_index)
            atom.SetNumExplicitHs(atom.GetNumExplicitHs() + 1)
            return edited_mol.GetMol()

        # Process a single component of the molecule
        def process_component(component):
            for atom_index in range(component.GetNumAtoms()):
                try:
                    test_mol = add_hydrogen(component, atom_index)
                    rdmolops.SanitizeMol(test_mol)
                    return test_mol
                except Exception:
                    continue
            return None

        components_smiles = smiles.split(".")
        valid_components = []

        for comp_smiles in components_smiles:
            comp = Chem.MolFromSmiles(comp_smiles, sanitize=False)
            valid_mol = process_component(comp)
            if valid_mol:
                valid_components.append(valid_mol)

        if valid_components:
            combined_mol = valid_components[0]
            for i in range(1, len(valid_components)):
                combined_mol = Chem.CombineMols(combined_mol, valid_components[i])
            return combined_mol
        else:
            return None



import unittest

from synrbl.SynMCSImputer.MissingGraph.uncertainty_graph import GraphMissingUncertainty


class TestGraphMissingUncertainty(unittest.TestCase):
    def setUp(self):
        self.graph_data = [
            {"boundary_atoms_products": [None, None], "smiles": ["C1=CC=CC=C1", None]},
            {
                "boundary_atoms_products": [{"C": 0}],
                "smiles": ["C1=CC=CC=C1.C2=CC=CC=C2"],
            },
            {"boundary_atoms_products": [], "smiles": ["C1=CC=CC=C1.C2=CC=CC=C2.O"]},
            {"boundary_atoms_products": [{"O": 2}], "smiles": ["O.O.O"]},
            {"boundary_atoms_products": [{"O": 2}], "smiles": ["O"]},
        ]
        self.threshold = 2
        self.graph_missing_uncertainty = GraphMissingUncertainty(
            self.graph_data, self.threshold
        )

    def test_check_boundary(self):
        without_boundary = GraphMissingUncertainty.check_boundary(self.graph_data)
        self.assertEqual(without_boundary, [0, 2])

    def test_check_fragments(self):
        graph_uncertain = GraphMissingUncertainty.check_fragments(
            self.graph_data, self.threshold
        )
        self.assertEqual(graph_uncertain, [1, 2, 3])

    def test_fit(self):
        updated_graph_data = self.graph_missing_uncertainty.fit()
        self.assertFalse(updated_graph_data[0]["Certainty"])
        self.assertFalse(updated_graph_data[1]["Certainty"])
        self.assertFalse(updated_graph_data[2]["Certainty"])
        self.assertFalse(updated_graph_data[3]["Certainty"])
        self.assertTrue(updated_graph_data[4]["Certainty"])


if __name__ == "__main__":
    unittest.main()



import re

from rdkit import Chem
from rdkit.Chem.rdMolDescriptors import CalcNumRings
from copy import deepcopy


def remove_atom_mapping_from_reaction_smiles(reaction_smiles):
    """
    Remove atom mapping from a reaction SMILES string.

    Parameters:
    - reaction_smiles (str): A reaction SMILES string with atom mapping.

    Returns:
    - str: A reaction SMILES string without atom mapping.
    """
    parts = reaction_smiles.split(">>")
    cleaned_parts = [Chem.CanonSmiles(re.sub(r":\d+", "", part)) for part in parts]
    cleaned_reaction_smiles = ">>".join(cleaned_parts)
    return cleaned_reaction_smiles


def calculate_chemical_properties(dictionary_list):
    """
    Calculate chemical properties from a list of dictionaries containing
    reaction information.

    Parameters:
    - dictionary_list (list of dict): A list of dictionaries with
        reaction information.

    Returns:
    - list of dict: Updated list with added chemical properties.
    """
    updated_list = deepcopy(dictionary_list)
    for entry in updated_list:
        reactant_smiles = entry["reactants"]
        product_smiles = entry["products"]
        reactant_mol = Chem.MolFromSmiles(reactant_smiles)
        product_mol = Chem.MolFromSmiles(product_smiles)

        if reactant_mol is not None and product_mol is not None:
            num_carbon_reactants = sum(
                [atom.GetAtomicNum() == 6 for atom in reactant_mol.GetAtoms()]
            )
            num_carbon_products = sum(
                [atom.GetAtomicNum() == 6 for atom in product_mol.GetAtoms()]
            )
            entry["carbon_difference"] = abs(num_carbon_reactants - num_carbon_products)
            entry["total_carbons"] = num_carbon_reactants + num_carbon_products
            entry["total_bonds"] = abs(
                reactant_mol.GetNumBonds() - product_mol.GetNumBonds()
            )
            entry["total_rings"] = abs(
                CalcNumRings(reactant_mol) - CalcNumRings(product_mol)
            )
        else:
            entry["carbon_difference"] = "Invalid SMILES"
            entry["total_carbons"] = "Invalid SMILES"
            entry["total_bonds"] = "Invalid SMILES"
            entry["total_rings"] = "Invalid SMILES"
        reactant_fragment_count = len(reactant_smiles.split("."))
        product_fragment_count = len(product_smiles.split("."))
        total_fragment_count = reactant_fragment_count + product_fragment_count
        entry["fragment_count"] = total_fragment_count

    return updated_list


def count_boundary_atoms_products_and_calculate_changes(
    list_of_dicts, reaction_col="new_reaction", mcs_col="mcs"
):
    """
    Count boundary atoms in products and calculate changes in bonds and rings.

    Parameters:
    - list_of_dicts (list of dict): A list of dictionaries with reaction information.

    Returns:
    - list of dict: Updated list with calculated changes.
    """
    for item in list_of_dicts:
        count = 0
        bond_change = 0
        ring_change = 0

        if mcs_col in item.keys():
            for i in item[mcs_col]["boundary_atoms_products"]:
                if isinstance(i, dict):
                    count += 1
                elif isinstance(i, list):
                    for j in i:
                        if isinstance(j, dict):
                            count += 1

        reactant_product = item[reaction_col].split(">>")
        if len(reactant_product) == 2:
            reactant_smiles, product_smiles = reactant_product
            reactant_mol = Chem.MolFromSmiles(reactant_smiles)
            product_mol = Chem.MolFromSmiles(product_smiles)

            if reactant_mol and product_mol:
                bond_change = abs(
                    reactant_mol.GetNumBonds() - product_mol.GetNumBonds()
                )
                ring_change = abs(
                    CalcNumRings(reactant_mol) - CalcNumRings(product_mol)
                )

        item["num_boundary"] = count
        item["bond_change_merge"] = bond_change
        item["ring_change_merge"] = ring_change

    return list_of_dicts



import io
import tempfile
import matplotlib.pyplot as plt

from typing import List, Dict
from PIL import Image
from reportlab.pdfgen import canvas
from reportlab.lib.pagesizes import landscape, letter
from synrbl.SynVis.reaction_visualizer import ReactionVisualizer


def save_reactions_to_pdf(
    data: List[Dict[str, str]],
    old_reaction_col: str,
    new_reaction_col: str,
    pdf_filename: str,
    compare: bool = False,
    orientation: str = "vertical",
    show_atom_numbers: bool = False,
):
    """
    Save a list of reaction visualizations to a PDF file.

    Parameters
    ----------
    data : List[Dict[str, str]]
        A list of dictionaries containing reaction data.
    old_reaction_col : str
        The column name containing the string representation of the old reaction.
    new_reaction_col : str
        The column name containing the string representation of the new reaction.
    pdf_filename : str
        The filename of the PDF to be saved.
    compare : bool, optional
        If True, both the old and new reactions are plotted. Default is False.
    orientation : str, optional
        The layout orientation of the plots ('vertical' or 'horizontal').
        Default is 'vertical'.
    show_atom_numbers : bool, optional
        Whether to show atom numbers in the reaction visualizations.
        Default is False.
    scale_factor : float, optional
        Factor to scale the reaction image size in the PDF. Default is 1.0.
    title_font_size : int, optional
        Font size for the title. Default is 14.

    Notes
    -----
    The method plots each reaction using the plot_reactions method and saves
    it to a PDF file. Each reaction is plotted on a separate page. The method
    also handles scaling of the reaction image and includes a customizable
    title for each reaction page in the PDF.
    """
    c = canvas.Canvas(pdf_filename, pagesize=landscape(letter))
    page_width, page_height = landscape(letter)

    for reaction_data in data:
        # Create a figure using plot_reactions method
        fig = ReactionVisualizer(figsize=(10, 5)).plot_reactions(
            reaction_data,
            old_reaction_col,
            new_reaction_col,
            compare,
            orientation,
            show_atom_numbers,
        )

        # Save figure to a temporary buffer
        buf = io.BytesIO()
        plt.savefig(buf, format="png")
        plt.close(fig)
        buf.seek(0)
        img = Image.open(buf)

        # Save the image to a temporary file
        with tempfile.NamedTemporaryFile(suffix=".png") as img_temp:
            img.save(img_temp, format="PNG")
            img_temp.flush()

            # Image dimensions
            img_width, img_height = img.size
            scale_factor = min(
                (page_width - 100) / img_width, (page_height - 100) / img_height
            )
            scaled_width, scaled_height = (
                img_width * scale_factor,
                img_height * scale_factor,
            )

            # Calculate coordinates to center the image
            x = (page_width - scaled_width) / 2
            y = (page_height - scaled_height) / 2

            # Draw the centered image
            c.drawImage(
                img_temp.name,
                x,
                y,
                width=scaled_width,
                height=scaled_height,
                mask="auto",
            )
            c.showPage()

    c.save()
    print(f"Saved reactions to {pdf_filename}")



import unittest

from synrbl.SynRuleImputer import RuleConstraint


class TestRuleConstraint(unittest.TestCase):
    def setUp(self):
        # Define sample reactions for testing
        self.sample_reactions = [
            {"reactants": "N.C=O", "products": "C=N.[O]"},
            {"reactants": "CO", "products": "C=O.[H].[H]"},
            {"reactants": "CBr.CBr", "products": "CC.BrBr"},
        ]
        self.ban_atoms = [
            "[O].[O]",
            "F-F",
            "Cl-Cl",
            "Br-Br",
            "I-I",
            "Cl-Br",
            "Cl-I",
            "Br-I",
        ]
        self.rule_constraint = RuleConstraint(self.sample_reactions, self.ban_atoms)

    def test_oxygen_products(self):
        # Expected outcome for the first reaction
        expected_outcome = {
            "reactants": "N.C=O.[H].[H]",
            "products": "C=N.O",
            "new_reaction": "N.C=O.[H].[H]>>C=N.O",
        }

        # Applying the fit method
        filtered_reactions, _ = self.rule_constraint.fit()

        # Assertions
        self.assertEqual(
            filtered_reactions[0]["new_reaction"], expected_outcome["new_reaction"]
        )

    def test_hydrogen_products(self):
        # Expected outcome for the second reaction
        expected_outcome = {
            "reactants": "CO.[O]",
            "products": "C=O.O",
            "new_reaction": "CO.[O]>>C=O.O",
        }

        # Applying the fit method
        filtered_reactions, _ = self.rule_constraint.fit()

        # Assertions
        self.assertEqual(
            filtered_reactions[1]["new_reaction"], expected_outcome["new_reaction"]
        )

    def test_reaction_banned_atoms(self):
        # Applying the fit method
        filtered_reactions, _ = self.rule_constraint.fit()

        # Assertions
        with self.assertRaises(IndexError):
            _ = filtered_reactions[2]


if __name__ == "__main__":
    unittest.main()



from typing import Dict
from rdkit import Chem


class AppelReaction:
    TCM = "ClC(Cl)(Cl)Cl"
    TCM_product = "ClC(Cl)Cl"
    TBM = "BrC(Br)(Br)Br"
    TBM_product = "BrC(Br)Br"
    TPP = "c1ccccc1P(c2ccccc2)c3ccccc3"
    TPPO = "O=P(c1ccccc1)(c2ccccc2)c3ccccc3"

    def __init__(self):
        pass

    @staticmethod
    def check_alcohol_group(smiles: str) -> bool:
        """
        Check for the presence of an alcohol (hydroxyl, -OH) group in a molecule.

        Args:
        smiles (str): A SMILES string representing the molecule.

        Returns:
        bool: True if the alcohol group is present, False otherwise.
        """
        alcohol_pattern = Chem.MolFromSmarts("[OH]")
        mol = Chem.MolFromSmiles(smiles)
        return mol.HasSubstructMatch(alcohol_pattern) if mol else False

    @staticmethod
    def check_appel_reaction(reactants: str) -> bool:
        """
        Check if the Appel reaction conditions are met in the reactants.

        Args:
        reactants (str): SMILES string of the reactants.

        Returns:
        bool: True if Appel reaction conditions are met, False otherwise.
        """
        return any(
            reagent in reactants for reagent in [AppelReaction.TCM, AppelReaction.TBM]
        ) and AppelReaction.check_alcohol_group(reactants)

    @staticmethod
    def check_missing_reagent(reactants: str) -> bool:
        """
        Check if the key reagent (Triphenylphosphine) is missing in the reactants.

        Args:
        reactants (str): SMILES string of the reactants.

        Returns:
        bool: True if Triphenylphosphine is missing, False otherwise.
        """
        return AppelReaction.TPP not in reactants

    @staticmethod
    def check_missing_products(products: str) -> bool:
        """
        Check if the expected product (Triphenylphosphine oxide) is missing in
        the products.

        Args:
        products (str): SMILES string of the products.

        Returns:
        bool: True if Triphenylphosphine oxide is missing, False otherwise.
        """
        return AppelReaction.TPPO not in products

    def fit(
        self, reaction_dict: Dict[str, str], rmsi_col: str, symbol: str = ">>"
    ) -> Dict[str, str]:
        """
        Modify the reaction dictionary to include missing reactants or products
        for Appel reaction.

        Args:
        reaction_dict (Dict[str, str]): Dictionary containing reaction information.
        rmsi_col (str): Key for the reaction SMILES string in the dictionary.
        symbol (str): Symbol used to separate reactants and products. Default is '>>'.

        Returns:
        Dict[str, str]: Updated reaction dictionary with modified SMILES string.
        """
        reaction = reaction_dict[rmsi_col]
        reactants, products = reaction.split(symbol)

        if self.check_appel_reaction(reactants):
            if self.check_missing_reagent(reactants):
                reactants += "." + AppelReaction.TPP
            if self.check_missing_products(products):
                products += "." + AppelReaction.TPPO

            # Add TCM or TBM products if they are missing
            if (
                AppelReaction.TCM in reactants
                and AppelReaction.TCM_product not in products
            ):
                products += "." + AppelReaction.TCM_product
            elif (
                AppelReaction.TBM in reactants
                and AppelReaction.TBM_product not in products
            ):
                products += "." + AppelReaction.TBM_product

            reaction_dict[rmsi_col] = symbol.join([reactants, products])

        return reaction_dict



import unittest

from synrbl.SynChemImputer.appel_reaction import (
    AppelReaction,
)


class TestAppelReaction(unittest.TestCase):
    def test_check_alcohol_group(self):
        # Test with a molecule containing an alcohol group
        self.assertTrue(AppelReaction.check_alcohol_group("CCO"))  # Ethanol

        # Test with a molecule not containing an alcohol group
        self.assertFalse(AppelReaction.check_alcohol_group("CCC"))  # Propane

    def test_check_appel_reaction(self):
        # Test with a valid Appel reaction
        self.assertTrue(
            AppelReaction.check_appel_reaction("ClC(Cl)(Cl)Cl.CCO")
        )  # TCM with ethanol

        # Test with an invalid Appel reaction
        self.assertFalse(AppelReaction.check_appel_reaction("CCO"))  # Only ethanol

    def test_check_missing_reagent(self):
        # Test with missing Triphenylphosphine
        self.assertTrue(AppelReaction.check_missing_reagent("ClC(Cl)(Cl)Cl.CCO"))

        # Test with present Triphenylphosphine
        self.assertFalse(
            AppelReaction.check_missing_reagent(
                "ClC(Cl)(Cl)Cl.CCO.c1ccccc1P(c2ccccc2)c3ccccc3"
            )
        )

    def test_check_missing_products(self):
        # Test with missing Triphenylphosphine oxide
        self.assertTrue(AppelReaction.check_missing_products("CCCl"))

        # Test with present Triphenylphosphine oxide
        self.assertFalse(
            AppelReaction.check_missing_products("CCCl.O=P(c1ccccc1)(c2ccccc2)c3ccccc3")
        )

    def test_fit(self):
        # Test the fit function with various scenarios
        appel_reactor = AppelReaction()
        reaction_dict = {"rmsi_col": "CCO.ClC(Cl)(Cl)Cl>>CCCl"}
        updated_dict = appel_reactor.fit(reaction_dict, "rmsi_col")
        self.assertIn(
            AppelReaction.TPPO, updated_dict["rmsi_col"]
        )  # Check for TPPO in products
        self.assertIn(
            AppelReaction.TCM_product, updated_dict["rmsi_col"]
        )  # Check for TCM product

        # Add more test cases as necessary

    def test_fit_missing_tp_po_and_tcm_product(self):
        appel_reactor = AppelReaction()
        reaction_dict = {"rmsi_col": "CCO.ClC(Cl)(Cl)Cl>>CCCl"}
        updated_dict = appel_reactor.fit(reaction_dict, "rmsi_col")
        self.assertIn(AppelReaction.TPPO, updated_dict["rmsi_col"])
        self.assertIn(AppelReaction.TCM_product, updated_dict["rmsi_col"])

    def test_fit_with_tcm_missing_tp_and_po(self):
        appel_reactor = AppelReaction()
        reaction_dict = {"rmsi_col": "CO.ClC(Cl)(Cl)Cl>>CCl"}
        updated_dict = appel_reactor.fit(reaction_dict, "rmsi_col")
        self.assertIn(AppelReaction.TPPO, updated_dict["rmsi_col"])

    def test_fit_with_existing_tcm_product(self):
        appel_reactor = AppelReaction()
        reaction_dict = {"rmsi_col": "CCO.ClC(Cl)(Cl)Cl>>CCCl.'ClC(Cl)Cl'"}
        updated_dict = appel_reactor.fit(reaction_dict, "rmsi_col")
        self.assertIn(AppelReaction.TPPO, updated_dict["rmsi_col"])

    def test_fit_with_existing_tp_po(self):
        appel_reactor = AppelReaction()
        reaction_dict = {
            "rmsi_col": "CCO.ClC(Cl)(Cl)Cl>>CCCl.'O=P(c1ccccc1)(c2ccccc2)c3ccccc3'"
        }
        updated_dict = appel_reactor.fit(reaction_dict, "rmsi_col")
        self.assertIn(AppelReaction.TPP, updated_dict["rmsi_col"])

    def test_fit_with_existing_tpp(self):
        appel_reactor = AppelReaction()
        reaction_dict = {
            "rmsi_col": "CCO.ClC(Cl)(Cl)Cl.c1ccccc1P(c2ccccc2)c3ccccc3>>CCCl'"
        }
        updated_dict = appel_reactor.fit(reaction_dict, "rmsi_col")
        self.assertIn(AppelReaction.TPPO, updated_dict["rmsi_col"])
        self.assertIn(AppelReaction.TCM_product, updated_dict["rmsi_col"])


if __name__ == "__main__":
    unittest.main()



import pandas as pd
import seaborn as sns
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors

from statsmodels.stats.proportion import proportion_confint
from matplotlib.ticker import AutoMinorLocator
from typing import List, Optional


class EDAVisualizer:
    """
    A class for visualizing accuracy metrics and confidence intervals of
    datasets across different categories, with support for line and bar
    charts, including error bars for confidence intervals. It dynamically
    adjusts subplot arrangements based on the number of metrics provided,
    allowing for a comprehensive and customizable visualization experience.

    Attributes
    ----------
    df : pd.DataFrame
        The dataset containing accuracy results and other metrics for
        visualization. It must contain a 'Result' column used for calculating
        accuracy and confidence intervals.
    columns : list of str
        Column names in the dataframe that represent different categories or
        groups for which accuracy and confidence intervals will be visualized.
    titles : list of str
        Custom titles for each subplot corresponding to the columns being
        visualized. These titles are used as x-axis labels for each chart.

    Methods
    -------
    calculate_accuracy_and_confidence(column):
        Calculates the accuracy and Wilson confidence intervals for a given
        column in the dataframe.

    visualize_accuracy(
        error_bar=True,
        chart_type="line",
        error_bar_color="black",
        same_color_scale=False,
        show_values=False,
        savepath=None,
    ):
        Generates visualizations for the accuracy metrics specified in the
        columns attribute, with optional error bars, supporting both line and
        bar chart types.
    """

    def __init__(self, df: pd.DataFrame, columns: List[str], titles: List[str]):
        """
        Initializes the AccuracyVisualizer with a dataframe, columns for
        visualization, and titles.

        Parameters
        ----------
        df : pd.DataFrame
            The dataset to visualize, containing at least one column for
            grouping and a 'Result' column for accuracy calculations.
        columns : list of str
            The columns in the dataset that should be visualized, representing
            different categories or groups.
        titles : list of str
            Titles for the subplots corresponding to each column, used as
            x-axis labels.
        """
        assert len(columns) == len(
            titles
        ), "Columns and titles must have the same length"
        self.df = df
        self.columns = columns
        self.titles = titles

    def calculate_accuracy_and_confidence(self, column: str) -> pd.DataFrame:
        """
        Calculates the accuracy and confidence intervals for a given column in
        the dataframe.

        This method groups the dataset by the specified column, calculates the
        sum and size for each group, determines the accuracy as the ratio of
        sum to size, and computes the confidence intervals using the Wilson
        method.

        Parameters
        ----------
        column : str
            The column for which to calculate accuracy and confidence
            intervals.

        Returns
        -------
        pd.DataFrame
            A dataframe with the specified column, accuracy, and confidence
            intervals (lower and upper bounds) for each group.
        """

        group_data = self.df.groupby(column)["Result"].agg(["sum", "size"])
        group_data["Accuracy"] = group_data["sum"] / group_data["size"]
        group_data = group_data[group_data["Accuracy"] > 0]

        confidence_lower, confidence_upper = proportion_confint(
            group_data["sum"], group_data["size"], method="wilson"
        )
        group_data["lower"] = group_data["Accuracy"] - confidence_lower
        group_data["upper"] = confidence_upper - group_data["Accuracy"]

        return group_data.reset_index()

    def visualize_accuracy(
        self,
        error_bar: bool = True,
        chart_type: str = "line",
        error_bar_color: str = "black",
        same_color_scale: bool = False,
        show_values: bool = False,
        save_path: Optional[str] = None,
    ) -> None:
        """
        Generates a visualization of the accuracy metrics specified in the
        columns attribute.

        This method supports dynamic adjustment of subplot layouts based on
        the number of metrics, optional error bars, and the choice between
        line and bar charts. It allows for a unified or varying color scale
        across subplots and the option to display numerical values on the
        charts.

        Parameters
        ----------
        error_bar : bool, optional
            Whether to include error bars for confidence intervals in the
            charts (default is True).
        chart_type : {'line', 'bar'}, optional
            The type of chart to use for visualization ('line' or 'bar',
            default is 'line').
        error_bar_color : str, optional
            Color for the error bars, if error_bar is True
            (default is 'black').
        same_color_scale : bool, optional
            Whether to use the same color scale across all subplots based on
            the maximum 'size' value in the dataset
            (default is False, which uses individual scales per subplot).
        show_values : bool, optional
            Whether to display numerical values on the charts
            (default is False).
        save_path : str or None, optional
            Path to save the figure. If None, the figure is not saved
            (default is None).

        Returns
        -------
        None
        """
        n_metrics = len(self.columns)
        if n_metrics == 1:
            nrows, ncols = 1, 1
        elif n_metrics == 2:
            nrows, ncols = 2, 1
        elif n_metrics == 3:
            nrows, ncols = 3, 1
        elif n_metrics == 4:
            nrows, ncols = 2, 2
        elif n_metrics in [5, 6]:
            nrows, ncols = 3, 2
        else:
            nrows, ncols = (n_metrics + 2) // 3, 3

        fig, axes = plt.subplots(
            nrows=nrows, ncols=ncols, figsize=(4.5 * ncols, 2.7 * nrows), squeeze=False
        )
        cmap = plt.get_cmap("flare", 256)
        labels = ["A", "B", "C", "D", "E", "F"]
        for i, (column, title) in enumerate(zip(self.columns, self.titles)):
            ax = axes.flatten()[i]
            data = self.calculate_accuracy_and_confidence(column)

            norm = (
                plt.Normalize(vmin=0, vmax=data["size"].max())
                if not same_color_scale
                else plt.Normalize(vmin=0, vmax=self.df["size"].max())
            )
            sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
            colors = [mcolors.to_rgba(cmap(norm(s))) for s in data["size"]]

            if chart_type == "line":
                sns.lineplot(
                    data=data,
                    x=column,
                    y="Accuracy",
                    marker="o",
                    ax=ax,
                    palette=colors,
                    color="black",
                    zorder=3,
                )
                if error_bar:
                    ax.errorbar(
                        data[column],
                        data["Accuracy"],
                        yerr=[data["lower"], data["upper"]],
                        fmt="o",
                        capsize=5,
                        ecolor="black",
                        elinewidth=1,
                    )
            elif chart_type == "bar":
                sns.barplot(
                    data=data, x=column, y="Accuracy", palette=colors, ax=ax, zorder=2
                )
                if error_bar:
                    x_positions = np.arange(len(data[column]))
                    ax.errorbar(
                        x=x_positions,
                        y=data["Accuracy"],
                        yerr=[data["lower"].values, data["upper"].values],
                        fmt="none",
                        ecolor="black",
                        elinewidth=1,
                    )
                if show_values:
                    for bar in ax.patches:
                        ax.annotate(
                            format(bar.get_height(), ".2f"),
                            (bar.get_x() + bar.get_width() / 2, bar.get_height()),
                            ha="center",
                            va="bottom",
                            xytext=(0, 5),
                            textcoords="offset points",
                        )

            ax.set_xlabel(title, labelpad=4, color="black")

            # Bring x-axis ticks and labels to the top
            ax.xaxis.tick_top()
            ax.xaxis.set_label_position(
                "bottom"
            )  # Set the x-axis label position to top

            # Adjust tick parameters for better visibility
            ax.tick_params(
                axis="x", which="major", direction="out", pad=2, colors="black"
            )
            ax.tick_params(axis="y", which="major", direction="out", colors="black")
            ax.text(
                -0.02,
                1.05,
                labels[i],
                transform=ax.transAxes,
                weight="bold",
                fontsize=12,
            )

            cbar_ax = plt.colorbar(sm, ax=ax, pad=0.05, aspect=10)
            cbar_ax.ax.tick_params(colors="black")
            if i % 2 == 0:
                ax.set_ylabel("Accuracy", color="black")
            else:
                ax.set_ylabel("")
                cbar_ax.ax.set_ylabel("Number of Samples", color="black")

            ax.set_ylim(0, 1)
            ax.yaxis.set_minor_locator(AutoMinorLocator(2))
            ax.grid(axis="y", which="major", linewidth=0.5, alpha=0.3)
            ax.grid(axis="y", which="minor", linewidth=0.1, alpha=0.3)

        plt.tight_layout()
        if save_path:
            plt.savefig(save_path, dpi=400, transparent=True)
        plt.show()



import copy

from rdkit import Chem
from synrbl.SynRuleImputer.synthetic_rule_matcher import SyntheticRuleMatcher
from joblib import Parallel, delayed
from typing import List, Dict, Any, Union, Optional
from rdkit import RDLogger

lg = RDLogger.logger()
lg.setLevel(RDLogger.ERROR)
RDLogger.DisableLog("rdApp.info")


class SyntheticRuleImputer(SyntheticRuleMatcher):
    """
    A class for imputing missing chemical data based on provided rules.

    Example:
    rule_dict = [{'smiles': 'C1=CC=CC=C1', 'Composition': {'C': 6, 'H': 6}},
                 {'smiles': 'CCO', 'Composition': {'C': 2, 'H': 6, 'O': 1}}]

    missing_data = [{'Diff_formula': {'C': 4, 'H': 4}, 'Unbalance': 'Products',
                     'reactants': 'C2H6', 'products': 'C2H6.C1=CC=CC=C1'},
                    {'Diff_formula': {'C': 3, 'H': 6}, 'Unbalance': 'Reactants',
                     'reactants': 'C2H6', 'products': 'C2H6.C1=CC=CC=C1'}]

    imputer = SyntheticRuleImputer(rule_dict, select='best', ranking='longest')
    imputed_data = imputer.impute(missing_data)

    print(imputed_data)
    # Output:
    # [{'Diff_formula': {'C': 4, 'H': 4}, 'Unbalance': 'Products',
    #   'reactants': 'C2H6', 'products': 'C2H6.C1=CC=CC=C1.CCO',
    #   'new_reaction': 'C2H6>>C2H6.C1=CC=CC=C1.CCO'},
    #  {'Diff_formula': {'C': 3, 'H': 6}, 'Unbalance': 'Reactants',
    #   'reactants': 'C2H6', 'products': 'C2H6.C1=CC=CC=C1',
    #   'new_reaction': 'C2H6>>C2H6.C1=CC=CC=C1'}]
    """

    def __init__(
        self,
        rule_dict: List[Dict[str, Any]],
        select: str = "best",
        ranking: str = "longest",
    ) -> None:
        """
        Initialize the SyntheticRuleImputer.

        Args:
            rule_dict: A list of dictionaries representing chemical rules.
            select: Selection mode, either 'best' for the best solution or
                'all' for all solutions.
            ranking: Ranking mode for solutions, options are 'longest',
                'least', 'greatest'.
        """
        self.rule_dict = rule_dict
        self.select = select
        self.ranking = ranking

    @staticmethod
    def single_impute(
        missing_dict: List[Dict[str, Any]],
        rule_dict: Dict[str, Any],
        select: str = "best",
        ranking: str = "longest",
    ) -> List[Dict[str, Any]]:
        """
        Impute missing chemical data based on the provided rules.

        Args:
            missing_dict (List[Dict[str, Any]]): A list of dictionaries
                representing missing chemical data.
            rule_dict (Dict[str, Any]): A dictionary representing the rules for
                imputing missing data.
            select (str, optional): The selection strategy for choosing the
                best rule. Defaults to 'best'.
            ranking (str, optional): The ranking strategy for ordering the
                rules. Defaults to 'longest'.

        Returns:
            List[Dict[str, Any]]: A list of dictionaries with imputed data.
        """
        dict_impute = copy.deepcopy(missing_dict)

        matcher = SyntheticRuleMatcher(
            rule_dict, dict_impute["Diff_formula"], select=select, ranking=ranking
        )
        solution = matcher.match()

        if solution:
            if len(solution[0]) > 0:
                valid_smiles = SyntheticRuleImputer.get_and_validate_smiles(solution[0])
                if valid_smiles:
                    key = (
                        "products"
                        if dict_impute["Unbalance"] == "Products"
                        else "reactants"
                    )
                    dict_impute[key] += "." + valid_smiles

                    # Construct the new_reaction key
                    dict_impute["new_reaction"] = (
                        dict_impute["reactants"] + ">>" + dict_impute["products"]
                    )

        return dict_impute

    def parallel_impute(self, missing_dict: List[Dict], n_jobs: int = 4) -> List[Dict]:
        """
        Impute missing chemical data in parallel.

        Args:
            missing_dict: A list of dictionaries representing missing chemical
                data.

        Returns:
            A list of dictionaries with imputed data.
        """
        dict_impute = Parallel(n_jobs=n_jobs, verbose=0)(
            delayed(self.single_impute)(item, self.rule_dict, self.select, self.ranking)
            for item in missing_dict
        )
        return dict_impute

    @staticmethod
    def get_and_validate_smiles(
        solution: List[Dict[str, Union[str, int]]]
    ) -> Optional[str]:
        """
        Concatenate smiles strings based on their ratios and validate the
        result using RDKit.

        Args:
            solution: A list of dictionaries representing chemical solutions,
                    each with a 'smiles' string and a 'Ratio' integer.

        Returns:
            A validated smiles string or None if validation fails.
        """
        # Concatenate smiles strings based on their ratios
        smiles_parts = []
        for item in solution:
            if "smiles" in item and "Ratio" in item:
                smiles_parts.extend([item["smiles"]] * item["Ratio"])
        new_smiles = ".".join(smiles_parts)

        # Validate the smiles string using RDKit
        if Chem.MolFromSmiles(new_smiles) is not None:
            return new_smiles
        return None



from .rsmi_comparator import RSMIComparator
from .rsmi_processing import RSMIProcessing
from .rsmi_decomposer import RSMIDecomposer
from .rsmi_both_side_process import BothSideReact
from .check_carbon_balance import CheckCarbonBalance
